import re

from pydantic import BaseModel, Field
from typing import List, Any, Optional, Dict, Tuple, Union, Callable
import collections
import math

from langchain.chains.llm import LLMChain
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_core.documents.base import Document
from langchain_core.embeddings import Embeddings
from langchain_core.exceptions import OutputParserException
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompts import PromptTemplate, ChatPromptTemplate
from langchain_core.retrievers import BaseRetriever

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    DistanceFunction,
    VectorStoreHandler,
)
from mindsdb.integrations.utilities.rag.settings import (
    DatabaseSchema,
    TableSchema,
    ColumnSchema,
    ValueSchema,
    SearchKwargs,
)
from mindsdb.utilities import log

import numpy as np

logger = log.getLogger(__name__)


class MetadataFilter(BaseModel):
    """Represents an LLM generated metadata filter to apply to a PostgreSQL query."""

    attribute: str = Field(description="Database column to apply filter to")
    comparator: str = Field(
        description="PostgreSQL comparator to use to filter database column"
    )
    value: Any = Field(description="Value to use to filter database column")


class AblativeMetadataFilter(MetadataFilter):
    """Adds additional fields to support ablation."""

    schema_table: str = Field(description="schema name of the table for this filter")
    schema_column: str = Field(description="schema name of the column for this filter")
    schema_value: str = Field(description="schema name of the value for this filter")


class MetadataFilters(BaseModel):
    """List of LLM generated metadata filters to apply to a PostgreSQL query."""

    filters: List[MetadataFilter] = Field(
        description="List of PostgreSQL metadata filters to apply for user query"
    )


class SQLRetriever(BaseRetriever):
    """Retriever that uses a LLM to generate pgvector queries to do similarity search with metadata filters.

    How it works:

    1. Use a LLM to rewrite the user input to something more suitable for retrieval. For example:
    "Show me documents containing how to finetune a LLM please" --> "how to finetune a LLM"

    2. Use a LLM to generate structured metadata filters based on the user input. Provided
    metadata schemas & examples are used as additional context.

    3. Generate a prepared PostgreSQL query from the structured metadata filters.

    4. Actually execute the query against our vector database to retrieve documents & return them.
    """

    fallback_retriever: BaseRetriever
    vector_store_handler: VectorStoreHandler
    # search parameters
    max_filters: int
    filter_threshold: float
    min_k: int

    # Schema description
    database_schema: Optional[DatabaseSchema] = None

    # Embeddings
    embeddings_model: Embeddings
    search_kwargs: SearchKwargs

    # prompt templates
    rewrite_prompt_template: str

    # schema templates
    table_prompt_template: str
    column_prompt_template: str
    value_prompt_template: str

    # formatting templates
    boolean_system_prompt: str
    generative_system_prompt: str

    # SQL search config
    num_retries: int
    embeddings_table: str
    source_table: str
    source_id_column: str
    distance_function: DistanceFunction

    # Re-rank and metadata generation model.
    llm: BaseChatModel

    def _sort_schema_by_priority_key(
        self,
        schema_dict_item: Tuple[str, Union[TableSchema, ColumnSchema, ValueSchema]],
    ):
        return schema_dict_item[1].priority

    def _sort_schema_by_relevance_key(
        self,
        schema_dict_item: Tuple[str, Union[TableSchema, ColumnSchema, ValueSchema]],
    ):
        if schema_dict_item[1].relevance is not None:
            return schema_dict_item[1].relevance
        else:
            return 0

    def _sort_schema_by_key(
        self,
        schema: Union[DatabaseSchema, TableSchema, ColumnSchema],
        key: Callable,
        update: Dict[str, Any] = None,
    ) -> Union[DatabaseSchema, TableSchema, ColumnSchema]:
        """Takes a schema and converts its dict into an OrderedDict"""
        if isinstance(schema, DatabaseSchema):
            collection_key = "tables"
        elif isinstance(schema, TableSchema):
            collection_key = "columns"
        elif isinstance(schema, ColumnSchema):
            collection_key = "values"
        else:
            raise Exception(
                "schema must be either a DatabaseSchema, TableSchema, or ColumnSchema."
            )

        if update is not None:
            ordered = collections.OrderedDict(
                sorted(update.items(), key=key, reverse=True)
            )
        else:
            ordered = collections.OrderedDict(
                sorted(getattr(schema, collection_key).items(), key=key, reverse=True)
            )
        schema = schema.model_copy(update={collection_key: ordered})

        return schema

    def _sort_database_schema_by_key(
        self, database_schema: DatabaseSchema, key: Callable
    ) -> DatabaseSchema:
        """Re-build schema with OrderedDicts"""
        tables = {}
        # build new tables dict
        for table_key, table_schema in database_schema.tables.items():
            columns = {}
            # build new column dict
            for column_key, column_schema in table_schema.columns.items():
                # sort values directly and update column schema
                columns[column_key] = self._sort_schema_by_key(
                    schema=column_schema, key=key
                )
            # update table schema and sort
            tables[table_key] = self._sort_schema_by_key(
                schema=table_schema, key=key, update=columns
            )
        # update table schema and sort
        database_schema = self._sort_schema_by_key(
            schema=database_schema, key=key, update=tables
        )

        return database_schema

    def _prepare_value_prompt(
        self,
        value_schema: ValueSchema,
        column_schema: ColumnSchema,
        table_schema: TableSchema,
        boolean_system_prompt: bool = True,
        format_instructions: Optional[str] = None,
    ) -> ChatPromptTemplate:

        if boolean_system_prompt is True:
            system_prompt = self.boolean_system_prompt
        else:
            system_prompt = self.generative_system_prompt

        prepared_column_prompt = self._prepare_column_prompt(
            column_schema=column_schema, table_schema=table_schema
        )
        column_schema_str = (
            prepared_column_prompt.messages[1]
            .format(
                **prepared_column_prompt.partial_variables,
                query="See query at the lowest level schema.",
            )
            .content
        )

        base_prompt_template = ChatPromptTemplate.from_messages(
            [("system", system_prompt), ("user", self.value_prompt_template)]
        )

        value_str = ""
        header_str = ""
        if type(value_schema.value) in [str, int, float, bool]:
            header_str = f"This schema describes a single value in the {column_schema.column} column."

            value_str = f"""
 -**Value**: {value_schema.value}
"""

        elif type(value_schema.value) is dict:
            header_str = f"This schema describes enumerated values in the {column_schema.column} column."

            value_str = """
## **Enumerated Values**

The values in the column are an enumeration of named values. These are listed below with format **[Column Value]**: [named value].
"""
            for value, value_name in value_schema.value.items():
                value_str += f"""
- **{value}:** {value_name}"""

        elif type(value_schema.value) is list:
            header_str = f"This schema describes some of the values in the {column_schema.column} column."

            value_str = """
## **Sample Values**

There are too many values in this column to list exhaustively. Below is a sampling of values found in the column:
"""
            for value in value_schema.value:
                value_str += f"""
- {value}"""

        if getattr(value_schema, "comparator", None) is not None:
            comparator_str = """

## **Comparators**

Below is a list of comparison operators for constructing filters for this value schema:
"""
            if type(value_schema.comparator) is str:
                comparator_str += f"""- {value_schema.comparator}
"""
            else:
                for comp in value_schema.comparator:
                    comparator_str += f"""- {comp}
"""
        else:
            comparator_str = ""

        if getattr(value_schema, "example_questions", None) is not None:
            example_str = """## **Example Questions**
"""
            for i, example in enumerate(value_schema.example_questions):
                example_str += f"""{i}. **Query:** {example.input} **Answer:** {example.output}
"""
        else:
            example_str = ""

        return base_prompt_template.partial(
            format_instructions=format_instructions,
            header=header_str,
            column_schema=column_schema_str,
            value=value_str,
            comparator=comparator_str,
            type=value_schema.type,
            description=value_schema.description,
            usage=value_schema.usage,
            examples=example_str,
        )

    def _prepare_column_prompt(
        self,
        column_schema: ColumnSchema,
        table_schema: TableSchema,
        boolean_system_prompt: bool = True,
    ) -> ChatPromptTemplate:

        if boolean_system_prompt is True:
            system_prompt = self.boolean_system_prompt
        else:
            system_prompt = self.generative_system_prompt

        prepared_table_prompt = self._prepare_table_prompt(
            table_schema=table_schema, boolean_system_prompt=boolean_system_prompt
        )
        table_schema_str = (
            prepared_table_prompt.messages[1]
            .format(
                **prepared_table_prompt.partial_variables,
                query="See query at the lowest level schema",
            )
            .content
        )

        base_prompt_template = ChatPromptTemplate.from_messages(
            [("system", system_prompt), ("user", self.column_prompt_template)]
        )

        header_str = (
            f"This schema describes a column in the {table_schema.table} table."
        )

        value_str = """
## **Content**

Below is a description of the contents in this column in list format:
"""
        for value_schema in column_schema.values.values():
            value_str += f"""
- {value_schema.description}
"""
        value_str += """
**Important:** The above descriptions are not the actual values stored in this column. See the Value schema for actual values.
"""

        if getattr(column_schema, "examples", None) is not None:
            example_str = """## **Example Questions**
"""
            for example in column_schema.examples:
                example_str += f"""- {example}
"""
        else:
            example_str = ""

        return base_prompt_template.partial(
            table_schema=table_schema_str,
            header=header_str,
            column=column_schema.column,
            type=column_schema.type,
            description=column_schema.description,
            usage=column_schema.usage,
            values=value_str,
            examples=example_str,
        )

    def _prepare_table_prompt(
        self, table_schema: TableSchema, boolean_system_prompt: bool = True
    ) -> ChatPromptTemplate:
        if boolean_system_prompt is True:
            system_prompt = self.boolean_system_prompt
        else:
            system_prompt = self.generative_system_prompt

        base_prompt_template = ChatPromptTemplate.from_messages(
            [("system", system_prompt), ("user", self.table_prompt_template)]
        )

        header_str = "This schema describes a table in the database."

        columns_str = ""
        for column_key, column_schema in table_schema.columns.items():
            columns_str += f"""
- **{column_schema.column}:** {column_schema.description}
"""

        if getattr(table_schema, "examples", None) is not None:
            example_str = """## **Example Questions**
"""
            for example in table_schema.examples:
                example_str += f"""- {example}
"""
        else:
            example_str = ""

        return base_prompt_template.partial(
            header=header_str,
            table=table_schema.table,
            description=table_schema.description,
            usage=table_schema.usage,
            columns=columns_str,
            examples=example_str,
        )

    def _rank_schema(self, prompt: ChatPromptTemplate, query: str) -> float:
        rank_chain = LLMChain(
            llm=self.llm.bind(logprobs=True), prompt=prompt, return_final_only=False
        )
        output = rank_chain({"query": query})  # returns metadata

        #  parse through metadata tokens until encountering either yes, or no.
        score = None  # a None score indicates the model output could not be parsed.
        for content in output["full_generation"][0].message.response_metadata[
            "logprobs"
        ]["content"]:
            #  Convert answer to score using the model's confidence
            if content["token"].lower().strip() == "yes":
                score = (
                    1 + math.exp(content["logprob"])
                ) / 2  # If yes, use the model's confidence
                break
            elif content["token"].lower().strip() == "no":
                score = (
                    1 - math.exp(content["logprob"])
                ) / 2  # If no, invert the confidence
                break

        if score is None:
            score = 0.0

        return score

    def _breadth_first_search(self, query: str, greedy: bool = False) -> Tuple:
        """Search breadth wise through Tables, then Columns, then Values.Uses a greedy strategy to maximize quota if greedy=True, otherwise a dynamic strategy."""

        # sort based on priority
        ordered_database_schema = self._sort_database_schema_by_key(
            database_schema=self.database_schema, key=self._sort_schema_by_priority_key
        )

        #  Rank Tables ########################################################
        greedy_count = 0
        tables = {}
        # rank tables by relevance
        for table_key, table_schema in ordered_database_schema.tables.items():
            prompt: ChatPromptTemplate = self._prepare_table_prompt(
                table_schema=table_schema, boolean_system_prompt=True
            )
            table_schema.relevance = self._rank_schema(prompt=prompt, query=query)

            # only keep greedy tables
            tables[table_key] = table_schema

            if greedy:
                if table_schema.relevance >= ordered_database_schema.filter_threshold:
                    greedy_count += 1
                if greedy_count >= ordered_database_schema.max_filters:
                    break

        #  sort tables
        ordered_database_schema = self._sort_schema_by_key(
            schema=ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        #  Rank Columns #######################################################
        #  iterate through tables to rank columns
        tables = {}
        table_count = 0  # take only the top n number of tables specified by the databases max filters
        for table_key, table_schema in ordered_database_schema.tables.items():
            # only drop into tables above the filter threshold
            if table_schema.relevance >= ordered_database_schema.filter_threshold:
                greedy_count = 0
                # rank columns by relevance
                columns = {}
                for column_key, column_schema in table_schema.columns.items():
                    prompt: ChatPromptTemplate = self._prepare_column_prompt(
                        column_schema=column_schema,
                        table_schema=table_schema,
                        boolean_system_prompt=True,
                    )
                    column_schema.relevance = self._rank_schema(
                        prompt=prompt, query=query
                    )

                    columns[column_key] = column_schema

                    if greedy:
                        if column_schema.relevance >= table_schema.filter_threshold:
                            greedy_count += 1
                        if greedy_count >= table_schema.max_filters:
                            break

                # sort columns and keep only columns that made the cut.
                tables[table_key] = self._sort_schema_by_key(
                    table_schema, key=self._sort_schema_by_relevance_key, update=columns
                )

                table_count += 1
                if table_count >= ordered_database_schema.max_filters:
                    break

        # sort tables and keep only tables that made the cut.
        ordered_database_schema = self._sort_schema_by_key(
            ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        #  Rank Values ########################################################
        #  iterate through tables to rank values
        tables = {}
        for table_key, table_schema in ordered_database_schema.tables.items():
            columns = {}
            column_count = 0
            # iterate through columns to rank values
            for column_key, column_schema in table_schema.columns.items():
                if column_schema.relevance >= table_schema.filter_threshold:
                    greedy_count = 0
                    values = {}
                    #  rank values by relevance
                    for value_key, value_schema in column_schema.values.items():
                        prompt: ChatPromptTemplate = self._prepare_value_prompt(
                            value_schema=value_schema,
                            column_schema=column_schema,
                            table_schema=table_schema,
                            boolean_system_prompt=True,
                        )
                        value_schema.relevance = self._rank_schema(
                            prompt=prompt, query=query
                        )

                        values[value_key] = value_schema

                        if greedy:
                            if value_schema.relevance >= column_schema.filter_threshold:
                                greedy_count += 1
                            if greedy_count >= column_schema.max_filters:
                                break

                    # sort values and keep only values that make the cut
                    columns[column_key] = self._sort_schema_by_key(
                        column_schema,
                        key=self._sort_schema_by_relevance_key,
                        update=values,
                    )

                    column_count += 1
                    if column_count >= table_schema.max_filters:
                        break

            # sort columns and keep only columns that made the cut
            tables[table_key] = self._sort_schema_by_key(
                table_schema, key=self._sort_schema_by_relevance_key, update=columns
            )

        # sort tables and keep only tables that made the cut.
        ordered_database_schema = self._sort_schema_by_key(
            ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        #  discard low ranked values ###################################################################################
        tables = {}
        for table_key, table_schema in ordered_database_schema.tables.items():
            columns = {}
            # iterate through columns to rank values
            for column_key, column_schema in table_schema.columns.items():
                value_count = 0
                values = {}
                #  rank values by relevance
                for value_key, value_schema in column_schema.values.items():
                    if value_schema.relevance >= column_schema.filter_threshold:
                        values[value_key] = value_schema

                        value_count += 1
                        if value_count >= column_schema.max_filters:
                            break

                # sort values and keep only values that make the cut
                columns[column_key] = self._sort_schema_by_key(
                    column_schema,
                    key=self._sort_schema_by_relevance_key,
                    update=values,
                )

            # sort columns and keep only columns that made the cut
            tables[table_key] = self._sort_schema_by_key(
                table_schema, key=self._sort_schema_by_relevance_key, update=columns
            )

        # sort tables and keep only tables that made the cut.
        ordered_database_schema = self._sort_schema_by_key(
            ordered_database_schema,
            key=self._sort_schema_by_relevance_key,
            update=tables,
        )

        ranked_database_schema = ordered_database_schema

        #  Build Ablation #####################################################

        ablation_value_dict = {}
        # assemble a relevance dictionary
        for table_key, table_schema in ordered_database_schema.tables.items():
            for column_key, column_schema in table_schema.columns.items():
                for value_key, value_schema in column_schema.values.items():
                    ablation_value_dict[(table_key, column_key, value_key)] = (
                        value_schema.relevance
                    )

        ablation_value_dict = collections.OrderedDict(
            sorted(ablation_value_dict.items(), key=lambda x: x[1])
        )

        relevance_scores = list(ablation_value_dict.values())
        if len(relevance_scores) > 0:
            ablation_quantiles = np.quantile(
                relevance_scores, np.linspace(0, 1, self.num_retries + 2)[1:-1]
            )
        else:
            ablation_quantiles = None

        return ranked_database_schema, ablation_value_dict, ablation_quantiles

    def _dynamic_ablation(
        self,
        metadata_filters: List[AblativeMetadataFilter],
        ablation_value_dict,
        ablation_quantiles,
        retry: int,
    ):
        """Ablate metadata filters in aggregate by quantiles until the required minimum number of documents are returned."""

        ablated_dict = {}
        for key, value in ablation_value_dict.items():
            if value >= ablation_quantiles[retry]:
                ablated_dict[key] = value

        #  discard low ranked filters ##################################################################################
        ablated_filters = []
        for filter in metadata_filters:
            for key in ablated_dict.keys():
                if (
                    filter.schema_table in key
                    and filter.schema_column in key
                    and filter.schema_value in key
                ):
                    ablated_filters.append(filter)

        return ablated_filters

    def depth_first_search(self, greedy=True):
        """Search depth wise through Tables, then Columns, then Values. Uses a greedy strategy to maximize quota if greedy=True, otherwise a dynamic strategy."""
        pass

    def depth_first_ablation(self):
        """Ablate metadata filters in reverse depth first search until the required minimum number of documents are returned."""
        pass

    def _prepare_retrieval_query(self, query: str) -> str:
        rewrite_prompt = PromptTemplate(
            input_variables=["input"], template=self.rewrite_prompt_template
        )
        rewrite_chain = LLMChain(llm=self.llm, prompt=rewrite_prompt)
        return rewrite_chain.predict(input=query)

    def _prepare_pgvector_query(
        self,
        ranked_database_schema: DatabaseSchema,
        metadata_filters: List[AblativeMetadataFilter],
        retry: int = 0,
    ) -> str:
        # Base select JOINed with document source table.
        base_query = f"""SELECT * FROM {self.embeddings_table} AS e INNER JOIN {self.source_table} AS s ON (e.metadata->>'original_row_id')::int = s."{self.source_id_column}" """

        # return an empty string if schema has not been ranked
        if not ranked_database_schema:
            return ""

        # Add Table JOIN statements
        join_clauses = set()
        for metadata_filter in metadata_filters:
            join_clause = ranked_database_schema.tables[
                metadata_filter.schema_table
            ].join
            if join_clause in join_clauses:
                continue
            else:
                join_clauses.add(join_clause)
                base_query += join_clause + " "

        # Add WHERE conditions from metadata filters
        if metadata_filters:
            base_query += "WHERE "
        for i, filter in enumerate(metadata_filters):
            value = filter.value
            if isinstance(value, str):
                value = f"'{value}'"
            base_query += f'"{filter.attribute}" {filter.comparator} {value}'
            if i < len(metadata_filters) - 1:
                base_query += " AND "

        base_query += f" ORDER BY e.embeddings {self.distance_function.value[0]} '{{embeddings}}' LIMIT {self.search_kwargs.k};"
        return base_query

    def _generate_filter(
        self, prompt: ChatPromptTemplate, query: str
    ) -> MetadataFilter:
        gen_filter_chain = LLMChain(llm=self.llm, prompt=prompt)
        output = gen_filter_chain({"query": query})
        return output

    def _generate_metadata_filters(
        self, query: str, ranked_database_schema
    ) -> Union[List[AblativeMetadataFilter], HandlerResponse]:
        parser = PydanticOutputParser(pydantic_object=MetadataFilter)

        metadata_filter_list = []
        #  iterate through tables to rank values
        for table_key, table_schema in ranked_database_schema.tables.items():
            # iterate through columns to rank values
            for column_key, column_schema in table_schema.columns.items():
                if column_schema.relevance >= table_schema.filter_threshold:
                    #  generate filters
                    for value_key, value_schema in column_schema.values.items():
                        # must use generation if field is a dictionary of tuples or a list
                        if type(value_schema.value) in [list, dict]:
                            try:
                                metadata_prompt: ChatPromptTemplate = (
                                    self._prepare_value_prompt(
                                        format_instructions=parser.get_format_instructions(),
                                        value_schema=value_schema,
                                        column_schema=column_schema,
                                        table_schema=table_schema,
                                        boolean_system_prompt=False,
                                    )
                                )

                                metadata_filters_chain = LLMChain(
                                    llm=self.llm, prompt=metadata_prompt
                                )
                                metadata_filter_output = metadata_filters_chain.predict(
                                    query=query,
                                )

                                # If the LLM outputs raw JSON, use it as-is.
                                # If the LLM outputs anything including a json markdown section, use the last one.
                                json_markdown_output = re.findall(
                                    r"```json.*```", metadata_filter_output, re.DOTALL
                                )
                                if json_markdown_output:
                                    metadata_filter_output = json_markdown_output[-1]
                                    # Clean the json tags.
                                    metadata_filter_output = metadata_filter_output[7:]
                                    metadata_filter_output = metadata_filter_output[:-3]

                                metadata_filter = parser.invoke(metadata_filter_output)
                                model_dump = metadata_filter.model_dump()
                                model_dump.update(
                                    {
                                        "schema_table": table_key,
                                        "schema_column": column_key,
                                        "schema_value": value_key,
                                    }
                                )
                                metadata_filter = AblativeMetadataFilter(**model_dump)
                            except OutputParserException as e:
                                logger.warning(
                                    f"LLM failed to generate structured metadata filters: {str(e)}"
                                )
                                return HandlerResponse(
                                    RESPONSE_TYPE.ERROR, error_message=str(e)
                                )
                        else:
                            metadata_filter = AblativeMetadataFilter(
                                attribute=column_schema.column,
                                comparator=value_schema.comparator,
                                value=value_schema.value,
                                schema_table=table_key,
                                schema_column=column_key,
                                schema_value=value_key,
                            )
                        metadata_filter_list.append(metadata_filter)

        return metadata_filter_list

    def _prepare_and_execute_query(
        self,
        ranked_database_schema: DatabaseSchema,
        metadata_filters: List[AblativeMetadataFilter],
        embeddings_str: str,
    ) -> HandlerResponse:
        try:
            checked_sql_query = self._prepare_pgvector_query(
                ranked_database_schema, metadata_filters
            )
            checked_sql_query_with_embeddings = checked_sql_query.format(
                embeddings=embeddings_str
            )
            return self.vector_store_handler.native_query(
                checked_sql_query_with_embeddings
            )
        except Exception as e:
            logger.warning(
                f"Failed to prepare and execute SQL query from structured metadata: {str(e)}"
            )
            return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        # Rewrite query to be suitable for retrieval.
        retrieval_query = self._prepare_retrieval_query(query)

        # Embed the rewritten retrieval query & include it in the similarity search pgvector query.
        embedded_query = self.embeddings_model.embed_query(retrieval_query)

        # Search for relevant filters
        ranked_database_schema, ablation_value_dict, ablation_quantiles = (
            self._breadth_first_search(query=query)
        )

        # Generate metadata filters
        metadata_filters = self._generate_metadata_filters(
            query=query, ranked_database_schema=ranked_database_schema
        )

        if type(metadata_filters) is list:
            # Initial Execution of the similarity search with metadata filters.
            document_response = self._prepare_and_execute_query(
                ranked_database_schema=ranked_database_schema,
                metadata_filters=metadata_filters,
                embeddings_str=str(embedded_query),
            )
            num_retries = 0
            while num_retries < self.num_retries:
                if (
                    document_response.resp_type != RESPONSE_TYPE.ERROR
                    and len(document_response.data_frame) >= self.min_k
                ):
                    # Successfully retrieved k documents to send to re-ranker.
                    break
                elif document_response.resp_type == RESPONSE_TYPE.ERROR:
                    # LLMs won't always generate structured metadata so we should have a fallback after retrying.
                    logger.info(
                        f"SQL Retriever query failed with error {document_response.error_message}"
                    )
                else:
                    logger.info(
                        f"SQL Retriever did not retrieve {self.min_k} documents: {len(document_response.data_frame)} documents retrieved."
                    )

                ablated_metadata_filters = self._dynamic_ablation(
                    metadata_filters=metadata_filters,
                    ablation_value_dict=ablation_value_dict,
                    ablation_quantiles=ablation_quantiles,
                    retry=num_retries,
                )

                document_response = self._prepare_and_execute_query(
                    ranked_database_schema=ranked_database_schema,
                    metadata_filters=ablated_metadata_filters,
                    embeddings_str=str(embedded_query),
                )

                num_retries += 1

            retrieved_documents = []
            if document_response.resp_type != RESPONSE_TYPE.ERROR:
                document_df = document_response.data_frame
                for _, document_row in document_df.iterrows():
                    retrieved_documents.append(
                        Document(
                            document_row.get("content", ""),
                            metadata=document_row.get("metadata", {}),
                        )
                    )
            if retrieved_documents:
                return retrieved_documents

            # If the SQL query constructed did not return any documents, fallback.
            logger.info(
                "No documents returned from SQL retriever, using fallback retriever."
            )
            return self.fallback_retriever._get_relevant_documents(
                retrieval_query, run_manager=run_manager
            )
        else:
            # If no metadata fields could be generated fallback.
            logger.info(
                "No metadata fields were successfully generated, using fallback retriever."
            )
            return self.fallback_retriever._get_relevant_documents(
                retrieval_query, run_manager=run_manager
            )



import time
import requests
from typing import Dict, List, Optional
from requests_toolbelt.multipart.encoder import MultipartEncoder

from mindsdb.utilities import log
from mindsdb.integrations.handlers.twelve_labs_handler.settings import twelve_labs_handler_config


logger = log.getLogger(__name__)


class TwelveLabsAPIClient:
    """
    The Twelve Labs API client for the Twelve Labs handler.
    This client is used for accessing the Twelve Labs API endpoints.
    """

    def __init__(self, api_key: str, base_url: str = None):
        """
        The initializer for the TwelveLabsAPIClient.

        Parameters
        ----------
        api_key : str
            The Twelve Labs API key.
        base_url : str, Optional
            The base URL for the Twelve Labs API. Defaults to the base URL in the Twelve Labs handler settings.
        """

        self.api_key = api_key
        self.headers = {
            'Content-Type': 'application/json',
            'x-api-key': self.api_key
        }
        self.base_url = base_url if base_url else twelve_labs_handler_config.BASE_URL

    def create_index(self, index_name: str, index_options: List[str], engine_id: Optional[str] = None, addons: Optional[List[str]] = None) -> str:
        """
        Create an index.

        Parameters
        ----------
        index_name : str
            Name of the index to be created.

        index_options : List[str]
            List of that specifies how the platform will process the videos uploaded to this index.

        engine_id : str, Optional
            ID of the engine. If not provided, the default engine is used.

        addons : List[str], Optional
            List of addons that should be enabled for the index.

        Returns
        -------
        str
            ID of the created index.
        """

        # TODO: change index_options to engine_options?
        # TODO: support multiple engines per index?
        body = {
            "index_name": index_name,
            "engines": [{
                "engine_name": engine_id if engine_id else twelve_labs_handler_config.DEFAULT_ENGINE,
                "engine_options": index_options
            }],
            "addons": addons,
        }

        result = self._submit_request(
            method="POST",
            endpoint="/indexes",
            data=body,
        )

        logger.info(f"Index {index_name} successfully created.")
        return result['_id']

    def get_index_by_name(self, index_name: str) -> str:
        """
        Get an index by name.

        Parameters
        ----------
        index_name : str
            Name of the index to be retrieved.

        Returns
        -------
        str
            ID of the index.
        """

        params = {
            "index_name": index_name,
        }

        result = self._submit_request(
            method="GET",
            endpoint="/indexes",
            data=params,
        )

        data = result['data']
        return data[0]['_id'] if data else None

    def list_videos_in_index(self, index_name: str) -> List[Dict]:
        """
        List videos in an index.

        Parameters
        ----------
        index_name : str
            Name of the index.

        Returns
        -------
        List[Dict]
            List of videos in the index.
        """

        index_id = self.get_index_by_name(index_name=index_name)

        data = []
        result = self._submit_request(
            method="GET",
            endpoint=f"/indexes/{index_id}/videos",
        )
        data.extend(result['data'])

        while result['page_info']['page'] < result['page_info']['total_page']:
            result = self._submit_request(
                method="GET",
                endpoint=f"/indexes/{index_id}/videos?page_token={result['page_info']['next_page_token']}",
            )
            data.extend(result['data'])

        logger.info(f"Retrieved videos in index {index_id} successfully.")

        return data

    def _update_video_metadata(self, index_id: str, video_id: str, video_title: str = None, metadata: Dict = None) -> None:
        """
        Update the metadata of a video that has already been indexed.

        Parameters
        ----------
        video_id : str
            ID of the video.

        video_title : str
            Title of the video.

        metadata : Dict, Optional
            Metadata to be updated.

        Returns
        -------
        None
        """

        body = {}

        if video_title:
            body['video_title'] = video_title

        if metadata:
            body['metadata'] = metadata

        self._submit_request(
            method="PUT",
            endpoint=f"/indexes/{index_id}/videos/{video_id}",
            data=body,
        )

        logger.info(f"Updated metadata for video {video_id} successfully.")

    def create_video_indexing_tasks(self, index_id: str, video_urls: List[str] = None, video_files: List[str] = None) -> List[str]:
        """
        Create video indexing tasks.

        Parameters
        ----------
        index_id : str
            ID of the index.

        video_urls : List[str], Optional
            List of video urls to be indexed. Either video_urls or video_files should be provided. This validation is handled by TwelveLabsHandlerModel.

        video_files : List[str], Optional
            List of video files to be indexed. Either video_urls or video_files should be provided. This validation is handled by TwelveLabsHandlerModel.

        Returns
        -------
        List[str]
            List of task IDs created.
        """

        task_ids = []

        if video_urls:
            logger.info("video_urls has been set, therefore, it will be given precedence.")
            logger.info("Creating video indexing tasks for video urls.")

            for video_url in video_urls:
                task_ids.append(
                    self._create_video_indexing_task(
                        index_id=index_id,
                        video_url=video_url
                    )
                )

        elif video_files:
            logger.info("video_urls has not been set, therefore, video_files will be used.")
            logger.info("Creating video indexing tasks for video files.")
            for video_file in video_files:
                task_ids.append(
                    self._create_video_indexing_task(
                        index_id=index_id,
                        video_file=video_file
                    )
                )

        return task_ids

    def _create_video_indexing_task(self, index_id: str, video_url: str = None, video_file: str = None) -> str:
        """
        Create a video indexing task.

        Parameters
        ----------
        index_id : str
            ID of the index.

        video_url : str, Optional
            URL of the video to be indexed. Either video_url or video_file should be provided. This validation is handled by TwelveLabsHandlerModel.

        video_file : str, Optional
            Path to the video file to be indexed. Either video_url or video_file should be provided. This validation is handled by TwelveLabsHandlerModel.

        Returns
        -------
        str
            ID of the created task.
        """

        body = {
            "index_id": index_id,
        }

        file_to_close = None
        if video_url:
            body['video_url'] = video_url

        elif video_file:
            import mimetypes
            # WE need the file open for the duration of the request. Maybe simplify it with context manager later, but needs _create_video_indexing_task re-written
            file_to_close = open(video_file, 'rb')
            mime_type, _ = mimetypes.guess_type(video_file)
            body['video_file'] = (file_to_close.name, file_to_close, mime_type)

        result = self._submit_multi_part_request(
            method="POST",
            endpoint="/tasks",
            data=body,
        )

        if file_to_close:
            file_to_close.close()

        task_id = result['_id']
        logger.info(f"Created video indexing task {task_id} for {video_url if video_url else video_file} successfully.")

        # update the video title
        video_reference = video_url if video_url else video_file
        task = self._get_video_indexing_task(task_id=task_id)
        self._update_video_metadata(
            index_id=index_id,
            video_id=task['video_id'],
            metadata={
                "video_reference": video_reference
            }
        )

        return task_id

    def poll_for_video_indexing_tasks(self, task_ids: List[str]) -> None:
        """
        Poll for video indexing tasks to complete.

        Parameters
        ----------
        task_ids : List[str]
            List of task IDs to be polled.

        Returns
        -------
        None
        """

        for task_id in task_ids:
            logger.info(f"Polling status of video indexing task {task_id}.")
            is_task_running = True

            while is_task_running:
                task = self._get_video_indexing_task(task_id=task_id)
                status = task['status']
                logger.info(f"Task {task_id} is in the {status} state.")

                wait_durtion = task['process']['remain_seconds'] if 'process' in task else twelve_labs_handler_config.DEFAULT_WAIT_DURATION

                if status in ('pending', 'indexing', 'validating'):
                    logger.info(f"Task {task_id} will be polled again in {wait_durtion} seconds.")
                    time.sleep(wait_durtion)

                elif status == 'ready':
                    logger.info(f"Task {task_id} completed successfully.")
                    is_task_running = False

                else:
                    logger.error(f"Task {task_id} failed with status {task['status']}.")
                    # TODO: update Exception to be more specific
                    raise Exception(f"Task {task_id} failed with status {task['status']}.")

        logger.info("All videos indexed successffully.")

    def _get_video_indexing_task(self, task_id: str) -> Dict:
        """
        Get a video indexing task.

        Parameters
        ----------
        task_id : str
            ID of the task.

        Returns
        -------
        Dict
            Video indexing task.
        """

        result = self._submit_request(
            method="GET",
            endpoint=f"/tasks/{task_id}",
        )

        logger.info(f"Retrieved video indexing task {task_id} successfully.")
        return result

    def search_index(self, index_id: str, query: str, search_options: List[str]) -> Dict:
        """
        Search an index.

        Parameters
        ----------
        index_id : str
            ID of the index.

        query : str
            Query to be searched.

        search_options : List[str]
            List of search options to be used.

        Returns
        -------
        Dict
            Search results.
        """

        body = {
            "index_id": index_id,
            "query": query,
            "search_options": search_options
        }

        data = []
        result = self._submit_request(
            method="POST",
            endpoint="/search",
            data=body,
        )
        data.extend(result['data'])

        while 'next_page_token' in result['page_info']:
            result = self._submit_request(
                method="GET",
                endpoint=f"/search/{result['page_info']['next_page_token']}"
            )
            data.extend(result['data'])

        logger.info(f"Search for index {index_id} completed successfully.")
        return data

    def summarize_videos(self, video_ids: List[str], summarization_type: str, prompt: str) -> Dict:
        """
        Summarize videos.

        Parameters
        ----------
        video_ids : List[str]
            List of video IDs.

        summarization_type : str
            Type of the summary to be generated. Supported types are 'summary', 'chapter' and 'highlight'.

        prompt: str
            Prompt to be used for the Summarize task

        Returns
        -------
        Dict
            Summary of the videos.
        """

        results = []
        results = [self.summarize_video(video_id, summarization_type, prompt) for video_id in video_ids]

        logger.info(f"Summarized videos {video_ids} successfully.")
        return results

    def summarize_video(self, video_id: str, summarization_type: str, prompt: str) -> Dict:
        """
        Summarize a video.

        Parameters
        ----------
        video_id : str
            ID of the video.

        summarization_type : str
            Type of the summary to be generated. Supported types are 'summary', 'chapter' and 'highlight'.

        prompt: str
            Prompt to be used for the Summarize task

        Returns
        -------
        Dict
            Summary of the video.
        """
        body = {
            "video_id": video_id,
            "type": summarization_type,
            "prompt": prompt
        }

        result = self._submit_request(
            method="POST",
            endpoint="/summarize",
            data=body,
        )

        logger.info(f"Video {video_id} summarized successfully.")
        return result

    def _submit_request(self, endpoint: str, headers: Dict = None, data: Dict = None, method: str = "GET") -> Dict:
        """
        Submit a request to the Twelve Labs API.

        Parameters
        ----------
        endpoint : str
            API endpoint.

        headers : Dict, Optional
            Headers to be used in the request.

        data : Dict, Optional
            Data to be used in the request.

        method : str, Optional
            HTTP method to be used in the request. Defaults to GET.

        Returns
        -------
        Dict
            Response from the API.
        """

        headers = headers if headers else self.headers

        if method == "GET":
            response = requests.get(
                url=self.base_url + endpoint,
                headers=headers,
                params=data if data else {},
            )

        elif method == "POST":
            response = requests.post(
                url=self.base_url + endpoint,
                headers=headers,
                json=data if data else {},
            )

        elif method == "PUT":
            response = requests.put(
                url=self.base_url + endpoint,
                headers=headers,
                json=data if data else {},
            )

        else:
            raise Exception(f"Method {method} not supported yet.")

        return self._handle_response(response)

    def _submit_multi_part_request(self, endpoint: str, headers: Dict = None, data: Dict = None, method: str = "POST") -> Dict:
        """
        Submit a multi-part request to the Twelve Labs API.

        Parameters
        ----------
        endpoint : str
            API endpoint.

        headers : Dict, Optional
            Headers to be used in the request.

        data : Dict, Optional
            Data to be used in the request.

        method : str, Optional
            HTTP method to be used in the request. Defaults to GET.

        Returns
        -------
        Dict
            Response from the API.
        """

        headers = headers = headers if headers else self.headers

        multipart_data = MultipartEncoder(fields=data)
        headers['Content-Type'] = multipart_data.content_type
        if method == "POST":
            response = requests.post(
                url=self.base_url + endpoint,
                headers=headers,
                data=multipart_data if multipart_data else {}
            )

        else:
            raise Exception(f"Method {method} not supported yet.")
        return self._handle_response(response)

    def _handle_response(self, response):
        if response.status_code in (200, 201):
            if response.content:
                result = response.json()
                logger.info("API request was successful.")
                return result
            else:
                logger.info("API request was successful. No content returned.")
                return {}
        else:
            if response.content:
                logger.error(f"API request has failed: {response.content}")
                # TODO: update Exception to be more specific
                raise Exception(f"API request has failed: {response.content}")
            else:
                logger.error("API request has failed. No content returned.")
                raise Exception("API request has failed. No content returned.")



import time
import threading
from typing import Optional, Callable
from concurrent.futures import ProcessPoolExecutor, Future

from pandas import DataFrame

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.ml_task_queue.const import ML_TASK_TYPE
from mindsdb.integrations.libs.ml_handler_process import (
    learn_process,
    update_process,
    predict_process,
    describe_process,
    create_engine_process,
    update_engine_process,
    create_validation_process,
    func_call_process
)


def init_ml_handler(module_path):
    import importlib  # noqa

    import mindsdb.integrations.libs.ml_handler_process  # noqa

    db.init()
    importlib.import_module(module_path)


def dummy_task():
    return None


def empty_callback(_task):
    return None


class MLProcessException(Exception):
    """Wrapper for exception to safely send it back to the main process.

    If exception can not be pickled (pickle.loads(pickle.dumps(e))) then it may lead to termination of the ML process.
    Also in this case, the error sent to the user will not be relevant. This wrapper should prevent it.
    """
    base_exception_bytes: bytes = None

    def __init__(self, base_exception: Exception, message: str = None) -> None:
        super().__init__(message)
        self.message = f'{base_exception.__class__.__name__}: {base_exception}'

    @property
    def base_exception(self) -> Exception:
        return RuntimeError(self.message)


class WarmProcess:
    """ Class-wrapper for a process that persist for a long time. The process
        may be initialized with any handler requirements. Current implimentation
        is based on ProcessPoolExecutor just because of multiprocessing.pool
        produce daemon processes, which can not be used for learning. That
        bahaviour may be changed only using inheritance.
    """
    def __init__(self, initializer: Optional[Callable] = None, initargs: tuple = ()):
        """ create and init new process

            Args:
                initializer (Callable): the same as ProcessPoolExecutor initializer
                initargs (tuple): the same as ProcessPoolExecutor initargs
        """
        self.pool = ProcessPoolExecutor(1, initializer=initializer, initargs=initargs)
        self.last_usage_at = time.time()
        self._markers = set()
        # region bacause of ProcessPoolExecutor does not start new process
        # untill it get a task, we need manually run dummy task to force init.
        self.task = self.pool.submit(dummy_task)
        self._init_done = False
        self.task.add_done_callback(self._init_done_callback)
        # endregion

    def __del__(self):
        self.shutdown()

    def shutdown(self, wait: bool = False) -> None:
        """Like ProcessPoolExecutor.shutdown

        Args:
            wait (bool): If True then shutdown will not return until all running futures have finished executing
        """
        self.pool.shutdown(wait=wait)

    def _init_done_callback(self, _task):
        """ callback for initial task
        """
        self._init_done = True

    def _update_last_usage_at_callback(self, _task):
        self.last_usage_at = time.time()

    def ready(self) -> bool:
        """ check is process ready to get a task or not

            Returns:
                bool
        """
        if self._init_done is False:
            self.task.result()
            self._init_done = True
        if self.task is None or self.task.done():
            return True
        return False

    def add_marker(self, marker: tuple):
        """ remember that that process processed task for that model

            Args:
                marker (tuple): identifier of model
        """
        if marker is not None:
            self._markers.add(marker)

    def has_marker(self, marker: tuple) -> bool:
        """ check if that process processed task for model

            Args:
                marker (tuple): identifier of model

            Returns:
                bool
        """
        if marker is None:
            return False
        return marker in self._markers

    def is_marked(self) -> bool:
        """ check if process has any marker

            Returns:
                bool
        """
        return len(self._markers) > 0

    def apply_async(self, func: Callable, *args: tuple, **kwargs: dict) -> Future:
        """ Run new task

            Args:
                func (Callable): function to run
                args (tuple): args to be passed to function
                kwargs (dict): kwargs to be passed to function

            Returns:
                Future
        """
        if not self.ready():
            raise Exception('Process task is not ready')
        self.task = self.pool.submit(
            func, *args, **kwargs
        )
        self.task.add_done_callback(self._update_last_usage_at_callback)
        self.last_usage_at = time.time()
        return self.task


def warm_function(func, context: str, *args, **kwargs):
    ctx.load(context)
    try:
        return func(*args, **kwargs)
    except Exception as e:
        if type(e) in (ImportError, ModuleNotFoundError):
            raise
        raise MLProcessException(base_exception=e)


class ProcessCache:
    """ simple cache for WarmProcess-es
    """
    def __init__(self, ttl: int = 120):
        """ Args:
            ttl (int) time to live for unused process
        """
        self.cache = {}
        self._init = False
        self._lock = threading.Lock()
        self._ttl = ttl
        self._keep_alive = {}
        self._stop_event = threading.Event()
        self.cleaner_thread = None

    def __del__(self):
        self._stop_clean()

    def _start_clean(self) -> None:
        """ start worker that close connections after ttl expired
        """
        if (
            isinstance(self.cleaner_thread, threading.Thread)
            and self.cleaner_thread.is_alive()
        ):
            return
        self._stop_event.clear()
        self.cleaner_thread = threading.Thread(target=self._clean, name='ProcessCache.clean')
        self.cleaner_thread.daemon = True
        self.cleaner_thread.start()

    def _stop_clean(self) -> None:
        """ stop clean worker
        """
        self._stop_event.set()

    def init(self):
        """ run processes for specified handlers
        """
        from mindsdb.interfaces.database.integrations import integration_controller
        preload_handlers = {}
        config = Config()
        is_cloud = config.get('cloud', False) # noqa

        if config['ml_task_queue']['type'] != 'redis':
            if is_cloud:
                lightwood_handler = integration_controller.get_handler_module('lightwood')
                if lightwood_handler is not None and lightwood_handler.Handler is not None:
                    preload_handlers[lightwood_handler.Handler] = 4 if is_cloud else 1

                huggingface_handler = integration_controller.get_handler_module('huggingface')
                if huggingface_handler is not None and huggingface_handler.Handler is not None:
                    preload_handlers[huggingface_handler.Handler] = 1

                openai_handler = integration_controller.get_handler_module('openai')
                if openai_handler is not None and openai_handler.Handler is not None:
                    preload_handlers[openai_handler.Handler] = 1

        with self._lock:
            if self._init is False:
                self._init = True
                for handler in preload_handlers:
                    self._keep_alive[handler.name] = preload_handlers[handler]
                    self.cache[handler.name] = {
                        'last_usage_at': time.time(),
                        'handler_module': handler.__module__,
                        'processes': [
                            WarmProcess(init_ml_handler, (handler.__module__,))
                            for _x in range(preload_handlers[handler])
                        ]
                    }

    def apply_async(self, task_type: ML_TASK_TYPE, model_id: Optional[int],
                    payload: dict, dataframe: Optional[DataFrame] = None) -> Future:
        """ run new task. If possible - do it in existing process, if not - start new one.

            Args:
                task_type (ML_TASK_TYPE): type of the task (learn, predict, etc)
                model_id (int): id of the model
                payload (dict): any 'lightweight' data that needs to be send in the process
                dataframe (DataFrame): DataFrame to be send in the process

            Returns:
                Future
        """
        self._start_clean()
        handler_module_path = payload['handler_meta']['module_path']
        integration_id = payload['handler_meta']['integration_id']
        if task_type in (ML_TASK_TYPE.LEARN, ML_TASK_TYPE.FINETUNE):
            func = learn_process
            kwargs = {
                'data_integration_ref': payload['data_integration_ref'],
                'problem_definition': payload['problem_definition'],
                'fetch_data_query': payload['fetch_data_query'],
                'project_name': payload['project_name'],
                'model_id': model_id,
                'base_model_id': payload.get('base_model_id'),
                'set_active': payload['set_active'],
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.PREDICT:
            func = predict_process
            kwargs = {
                'predictor_record': payload['predictor_record'],
                'ml_engine_name': payload['handler_meta']['engine'],
                'args': payload['args'],
                'dataframe': dataframe,
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.DESCRIBE:
            func = describe_process
            kwargs = {
                'attribute': payload.get('attribute'),
                'model_id': model_id,
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.CREATE_VALIDATION:
            func = create_validation_process
            kwargs = {
                'target': payload.get('target'),
                'args': payload.get('args'),
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.CREATE_ENGINE:
            func = create_engine_process
            kwargs = {
                'connection_args': payload['connection_args'],
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.UPDATE_ENGINE:
            func = update_engine_process
            kwargs = {
                'connection_args': payload['connection_args'],
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.UPDATE:
            func = update_process
            kwargs = {
                'args': payload['args'],
                'integration_id': integration_id,
                'model_id': model_id,
                'module_path': handler_module_path
            }
        elif task_type == ML_TASK_TYPE.FUNC_CALL:
            func = func_call_process
            kwargs = {
                'name': payload['name'],
                'args': payload['args'],
                'integration_id': integration_id,
                'module_path': handler_module_path
            }
        else:
            raise Exception(f'Unknown ML task type: {task_type}')

        ml_engine_name = payload['handler_meta']['engine']
        model_marker = (model_id, payload['context']['company_id'])
        with self._lock:
            if ml_engine_name not in self.cache:
                warm_process = WarmProcess(init_ml_handler, (handler_module_path,))
                self.cache[ml_engine_name] = {
                    'last_usage_at': None,
                    'handler_module': handler_module_path,
                    'processes': [warm_process]
                }
            else:
                warm_process = None
                if model_marker is not None:
                    try:
                        warm_process = next(
                            p for p in self.cache[ml_engine_name]['processes']
                            if p.ready() and p.has_marker(model_marker)
                        )
                    except StopIteration:
                        pass
                if warm_process is None:
                    try:
                        warm_process = next(
                            p for p in self.cache[ml_engine_name]['processes']
                            if p.ready()
                        )
                    except StopIteration:
                        pass
                if warm_process is None:
                    warm_process = WarmProcess(init_ml_handler, (handler_module_path,))
                    self.cache[ml_engine_name]['processes'].append(warm_process)

            task = warm_process.apply_async(warm_function, func, payload['context'], **kwargs)
            self.cache[ml_engine_name]['last_usage_at'] = time.time()
            warm_process.add_marker(model_marker)
        return task

    def _clean(self) -> None:
        """ worker that stop unused processes
        """
        while self._stop_event.wait(timeout=10) is False:
            with self._lock:
                for handler_name in self.cache.keys():
                    processes = self.cache[handler_name]['processes']
                    processes.sort(key=lambda x: x.is_marked())

                    expected_count = 0
                    if handler_name in self._keep_alive:
                        expected_count = self._keep_alive[handler_name]

                    # stop processes which was used, it needs to free memory
                    for i, process in enumerate(processes):
                        if (
                            process.ready()
                            and process.is_marked()
                            and (time.time() - process.last_usage_at) > self._ttl
                        ):
                            processes.pop(i)
                            # del process
                            process.shutdown()
                            break

                    while expected_count > len(processes):
                        processes.append(
                            WarmProcess(init_ml_handler, (self.cache[handler_name]['handler_module'],))
                        )

    def shutdown(self, wait: bool = True) -> None:
        """Call 'shutdown' for each process cache

        wait (bool): like ProcessPoolExecutor.shutdown wait arg.
        """
        with self._lock:
            for handler_name in self.cache:
                for process in self.cache[handler_name]['processes']:
                    process.shutdown(wait=wait)
                self.cache[handler_name]['processes'] = []

    def remove_processes_for_handler(self, handler_name: str) -> None:
        """
            Remove all warm processes for a given handler.
            This is useful when the previous processes use an outdated instance of the handler.
            A good example is when the dependencies for a handler are installed after attempting to use the handler.

            Args:
                handler_name (str): name of the handler.
        """
        with self._lock:
            if handler_name in self.cache:
                for process in self.cache[handler_name]['processes']:
                    process.shutdown()

                self.cache[handler_name]['processes'] = []


process_cache = ProcessCache()



import os
import shutil
import tempfile

import pandas as pd
from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import CreateTable, DropTables, Insert, Select, Identifier
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.api.executor.utilities.sql import query_df
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log


logger = log.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 250


def clean_cell(val):
    if str(val) in ["", " ", "  ", "NaN", "nan", "NA"]:
        return None
    return val


class FileHandler(DatabaseHandler):
    """
    Handler for files
    """

    name = "files"

    def __init__(
        self,
        name=None,
        file_storage=None,
        connection_data={},
        file_controller=None,
        **kwargs,
    ):
        super().__init__(name)
        self.parser = parse_sql
        self.fs_store = file_storage
        self.custom_parser = connection_data.get("custom_parser", None)
        self.clean_rows = connection_data.get("clean_rows", True)
        self.chunk_size = connection_data.get("chunk_size", DEFAULT_CHUNK_SIZE)
        self.chunk_overlap = connection_data.get("chunk_overlap", DEFAULT_CHUNK_OVERLAP)
        self.file_controller = file_controller

    def connect(self, **kwargs):
        return

    def disconnect(self, **kwargs):
        return

    def check_connection(self) -> StatusResponse:
        return StatusResponse(True)

    def _get_table_page_names(self, table: Identifier):
        table_name_parts = table.parts

        # Check if it's a multi-part name (e.g., `file_name.sheet_name`)
        if len(table_name_parts) > 1:
            table_name = table_name_parts[-2]
            page_name = table_name_parts[-1]  # Get the sheet name
        else:
            table_name = table_name_parts[-1]
            page_name = None
        return table_name, page_name

    def query(self, query: ASTNode) -> Response:
        if type(query) is DropTables:
            for table_identifier in query.tables:
                if (
                    len(table_identifier.parts) == 2
                    and table_identifier.parts[0] != self.name
                ):
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table from database '{table_identifier.parts[0]}'",
                    )
                table_name = table_identifier.parts[-1]
                try:
                    self.file_controller.delete_file(table_name)
                except Exception as e:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Can't delete table '{table_name}': {e}",
                    )
            return Response(RESPONSE_TYPE.OK)

        if isinstance(query, CreateTable):
            # Check if the table already exists or if the table name contains more than one namespace
            existing_files = self.file_controller.get_files_names()

            if len(query.name.parts) != 1:
                return Response(
                    RESPONSE_TYPE.ERROR,
                    error_message="Table name cannot contain more than one namespace",
                )

            table_name = query.name.parts[-1]
            if table_name in existing_files:
                if query.is_replace:
                    self.file_controller.delete_file(table_name)
                else:
                    return Response(
                        RESPONSE_TYPE.ERROR,
                        error_message=f"Table '{table_name}' already exists",
                    )

            temp_dir_path = tempfile.mkdtemp(prefix="mindsdb_file_")

            try:
                # Create a temp file to save the table
                temp_file_path = os.path.join(temp_dir_path, f"{table_name}.csv")

                # Create an empty file using with the columns in the query
                df = pd.DataFrame(columns=[col.name for col in query.columns])
                df.to_csv(temp_file_path, index=False)

                self.file_controller.save_file(table_name, temp_file_path, file_name=f"{table_name}.csv")
            except Exception as unknown_error:
                return Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=f"Error creating table '{table_name}': {unknown_error}",
                )
            finally:
                # Remove the temp dir created
                shutil.rmtree(temp_dir_path, ignore_errors=True)

            return Response(RESPONSE_TYPE.OK)

        elif isinstance(query, Select):
            table_name, page_name = self._get_table_page_names(query.from_table)

            df = self.file_controller.get_file_data(table_name, page_name)

            # Process the SELECT query
            result_df = query_df(df, query)
            return Response(RESPONSE_TYPE.TABLE, data_frame=result_df)

        elif isinstance(query, Insert):
            table_name, page_name = self._get_table_page_names(query.table)

            df = self.file_controller.get_file_data(table_name, page_name)

            # Create a new dataframe with the values from the query
            new_df = pd.DataFrame(query.values, columns=[col.name for col in query.columns])

            # Concatenate the new dataframe with the existing one
            df = pd.concat([df, new_df], ignore_index=True)

            self.file_controller.set_file_data(table_name, df, page_name=page_name)

            return Response(RESPONSE_TYPE.OK)

        else:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message="Only 'select', 'insert', 'create' and 'drop' queries allowed for files",
            )

    def native_query(self, query: str) -> Response:
        ast = self.parser(query)
        return self.query(ast)

    def get_tables(self) -> Response:
        """
        List all files
        """
        files_meta = self.file_controller.get_files()
        data = [
            {
                "TABLE_NAME": x["name"],
                "TABLE_ROWS": x["row_count"],
                "TABLE_TYPE": "BASE TABLE",
            }
            for x in files_meta
        ]
        return Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(data))

    def get_columns(self, table_name) -> Response:
        file_meta = self.file_controller.get_file_meta(table_name)
        result = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                [
                    {
                        "Field": x["name"].strip()
                        if isinstance(x, dict)
                        else x.strip(),
                        "Type": "str",
                    }
                    for x in file_meta["columns"]
                ]
            ),
        )
        return result



import unittest

from mindsdb.integrations.handlers.couchbase_handler.couchbase_handler import (
    CouchbaseHandler,
)
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class CouchbaseHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        connection_data = {
            "host": "192.168.33.10",
            "user": "admin",
            "password": "00154abs",
            "bucket": "bag-bucket",
            "scope": "test-scope",  # This is optinal, but if ommited will default to _default.
        }
        cls.kwargs = dict(connection_data=connection_data)
        cls.handler = CouchbaseHandler("test_couchbase_handler", **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_1_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_2_get_column(self):
        tbls = self.handler.get_columns("onsale")
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_3_native_query_select(self):
        tbls = self.handler.native_query("SELECT * FROM onsale")
        assert tbls.type is not RESPONSE_TYPE.ERROR


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import List, Union, Any, Optional, Dict, OrderedDict

from langchain_community.vectorstores.chroma import Chroma
from langchain_community.vectorstores.pgvector import PGVector
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_core.vectorstores import VectorStore
from langchain_core.stores import BaseStore
from pydantic import BaseModel, Field, field_validator
from langchain_text_splitters import TextSplitter

DEFAULT_COLLECTION_NAME = "default_collection"

# Multi retriever specific
DEFAULT_ID_KEY = "doc_id"
DEFAULT_MAX_CONCURRENCY = 5
DEFAULT_K = 20

DEFAULT_CARDINALITY_THRESHOLD = 40
DEFAULT_MAX_SUMMARIZATION_TOKENS = 4000
DEFAULT_CHUNK_SIZE = 1000
DEFAULT_CHUNK_OVERLAP = 200
DEFAULT_POOL_RECYCLE = 3600
DEFAULT_LLM_MODEL = "gpt-4o"
DEFAULT_LLM_MODEL_PROVIDER = "openai"
DEFAULT_CONTENT_COLUMN_NAME = "body"
DEFAULT_DATASET_DESCRIPTION = "email inbox"
DEFAULT_TEST_TABLE_NAME = "test_email"
DEFAULT_VECTOR_STORE = Chroma
DEFAULT_RERANKER_FLAG = False
DEFAULT_RERANKING_MODEL = "gpt-4o"
DEFAULT_LLM_ENDPOINT = "https://api.openai.com/v1"
DEFAULT_AUTO_META_PROMPT_TEMPLATE = """
Below is a json representation of a table with information about {description}.
Return a JSON list with an entry for each column. Each entry should have
{{"name": "column name", "description": "column description", "type": "column data type"}}
\n\n{dataframe}\n\nJSON:\n
"""
DEFAULT_RAG_PROMPT_TEMPLATE = """You are an assistant for
question-answering tasks. Use the following pieces of retrieved context
to answer the question. If you don't know the answer, just say that you
don't know. Use two sentences maximum and keep the answer concise.
Question: {question}
Context: {context}
Answer:"""

DEFAULT_QA_GENERATION_PROMPT_TEMPLATE = """You are an assistant for
generating sample questions and answers from the given document and metadata. Given
a document and its metadata as context, generate a question and answer from that document and its metadata.

The document will be a string. The metadata will be a JSON string. You need
to parse the JSON to understand it.

Generate a question that requires BOTH the document and metadata to answer, if possible.
Otherwise, generate a question that requires ONLY the document to answer.

Return a JSON dictionary with the question and answer like this:
{{ "question": <the full generated question>, "answer": <the full generated answer> }}

Make sure the JSON string is valid before returning it. You must return the question and answer
in the specified JSON format no matter what.

Document: {document}
Metadata: {metadata}
Answer:"""

DEFAULT_MAP_PROMPT_TEMPLATE = """The following is a set of documents
{docs}
Based on this list of docs, please summarize based on the user input.

User input: {input}

Helpful Answer:"""

DEFAULT_REDUCE_PROMPT_TEMPLATE = """The following is set of summaries:
{docs}
Take these and distill it into a final, consolidated summary related to the user input.

User input: {input}

Helpful Answer:"""

DEFAULT_SEMANTIC_PROMPT_TEMPLATE = """Provide a better search query for web search engine to answer the given question.

<< EXAMPLES >>
1. Input: "Show me documents containing how to finetune a LLM please"
Output: "how to finetune a LLM"

Output only a single better search query and nothing else like in the example.

Here is the user input: {input}
"""

DEFAULT_METADATA_FILTERS_PROMPT_TEMPLATE = """Construct a list of PostgreSQL metadata filters to filter documents in the database based on the user input.

<< INSTRUCTIONS >>
{format_instructions}

RETURN ONLY THE FINAL JSON. DO NOT EXPLAIN, JUST RETURN THE FINAL JSON.

<< TABLES YOU HAVE ACCESS TO >>

{schema}

<< EXAMPLES >>

{examples}

Here is the user input:
{input}
"""

DEFAULT_BOOLEAN_PROMPT_TEMPLATE = """**Task:** Determine Schema Relevance for Database Search Queries

As an expert in constructing database search queries, you are provided with database schemas detailing tables, columns, and values. Your task is to assess whether these elements can be used to effectively search the database in relation to a given user query.

**Instructions:**

- **Evaluate the Schema**:
  - Analyze the tables, columns, and values described.
  - Consider their potential usefulness in retrieving information pertinent to the user query.

- **Decision Criteria**:
  - Determine if any part of the schema could assist in forming a relevant search query for the information requested.

- **Response**:
  - Reply with a single word: 'yes' if the schema components are useful, otherwise 'no'.

**Note:** Provide your answer based solely on the relevance of the described schema to the user query."""

DEFAULT_GENERATIVE_SYSTEM_PROMPT = """You are an expert database analyst that can assist in building SQL queries by providing structured output. Follow these format instructions precisely to generate a metadata filter given the provided schema description.

## Format instructions:
{format_instructions}
 """

DEFAULT_VALUE_PROMPT_TEMPLATE = """
{column_schema}

# **Value Schema**
{header}

- The type of the value: {type}

## **Description**
{description}

{value}{comparator}

## **Usage**
{usage}

{examples}

## **Query**
{query}

"""

DEFAULT_COLUMN_PROMPT_TEMPLATE = """
{table_schema}

# **Column Schema**
{header}

- The column name in the database table: {column}
- The type of the values in this column: {type}

## **Description**
{description}

## **Usage**
{usage}

{examples}

## **Query**
{query}
"""

DEFAULT_TABLE_PROMPT_TEMPLATE = """# **Table Schema**
{header}

- The name of this table in the database: {table}

## **Description**
{description}

## **Usage**
{usage}

## **Column Descriptions**
Below are descriptions of each column in this table:

{columns}

{examples}

## **Query**
{query}
"""

DEFAULT_SQL_PROMPT_TEMPLATE = """
Construct a valid {dialect} SQL query to select documents relevant to the user input.
Source documents are found in the {source_table} table. You may need to join with other tables to get additional document metadata.

The JSON col "metadata" in the {embeddings_table} has a string field called "original_row_id". This "original_row_id" string field in the
"metadata" col is the document ID associated with a row in the {embeddings_table} table.
You MUST always join with the {embeddings_table} table containing vector embeddings for the documents. For example, for a table named sd with an id column "Id":
JOIN {embeddings_table} v ON (v."metadata"->>'original_row_id')::int = sd."Id"

You MUST always order the embeddings by the {distance_function} comparator with '{{embeddings}}'.
You MUST always limit by {k} returned documents.
For example:
ORDER BY v.embeddings {distance_function} '{{embeddings}}' LIMIT {k};


<< TABLES YOU HAVE ACCESS TO >>
1. {embeddings_table} - Contains document chunks, vector embeddings, and metadata for documents.
You MUST always include the metadata column in your SELECT statement.
You MUST always join with the {embeddings_table} table containing vector embeddings for the documents.
You MUST always order by the provided embeddings vector using the {distance_function} comparator.
You MUST always limit by {k} returned documents.

Columns:
```json
{{
    "id": {{
        "type": "string",
        "description": "Unique ID for this document chunk"
    }},
    "content": {{
        "type": "string",
        "description": "A document chunk (subset of the original document)"
    }},
    "embeddings": {{
        "type": "vector",
        "description": "Vector embeddings for the document chunk. ALWAYS order by the provided embeddings vector using the {distance_function} comparator."
    }},
    "metadata": {{
        "type": "jsonb",
        "description": "Metadata for the document chunk. Always select metadata and always join with the {source_table} table on the string metadata field 'original_row_id'"
    }}
}}

{schema}

<< EXAMPLES >>

{examples}

Output the {dialect} SQL query that is ready to be executed only WITHOUT ANY DELIMITERS. Make sure to properly quote identifiers.

Here is the user input:
{input}
"""

DEFAULT_QUESTION_REFORMULATION_TEMPLATE = """Given the original question and the retrieved context,
analyze what additional information is needed for a complete, accurate answer.

Original Question: {question}

Retrieved Context:
{context}

Analysis Instructions:
1. Evaluate Context Coverage:
   - Identify key entities and concepts from the question
   - Check for temporal information (dates, periods, sequences)
   - Verify causal relationships are explained
   - Confirm presence of requested quantitative data
   - Assess if geographic or spatial context is sufficient

2. Quality Assessment:
   If the retrieved context is:
   - Irrelevant or tangential
   - Too general or vague
   - Potentially contradictory
   - Missing key perspectives
   - Lacking proper evidence
   Generate questions to address these specific gaps.

3. Follow-up Question Requirements:
   - Questions must directly contribute to answering the original query
   - Break complex relationships into simpler, sequential steps
   - Maintain specificity rather than broad inquiries
   - Avoid questions answerable from existing context
   - Ensure questions build on each other logically
   - Limit questions to 150 characters each
   - Each question must be self-contained
   - Questions must end with a question mark

4. Response Format:
   - Return a JSON array of strings
   - Use square brackets and double quotes
   - Questions must be unique (no duplicates)
   - If context is sufficient, return empty array []
   - Maximum 3 follow-up questions
   - Minimum length per question: 30 characters
   - No null values or empty strings

Example:
Original: "How did the development of antibiotics affect military casualties in WWII?"

Invalid responses:
{'questions': ['What are antibiotics?']}  // Wrong format
['What is WWII?']  // Too basic
['How did it impact things?']  // Too vague
['', 'Question 2']  // Contains empty string
['Same question?', 'Same question?']  // Duplicate

Valid response:
["What were military casualty rates from infections before widespread antibiotic use in 1942?",
 "How did penicillin availability change throughout different stages of WWII?",
 "What were the primary battlefield infections treated with antibiotics during WWII?"]

or [] if context fully answers the original question.

Your task: Based on the analysis of the original question and context,
output ONLY a JSON array of follow-up questions needed to provide a complete answer.
If no additional information is needed, output an empty array [].

Follow-up Questions:"""

DEFAULT_QUERY_RETRY_PROMPT_TEMPLATE = """
{query}

The {dialect} query above failed with the error message: {error}.

<< TABLES YOU HAVE ACCESS TO >>
1. {embeddings_table} - Contains document chunks, vector embeddings, and metadata for documents.

Columns:
```json
{{
    "id": {{
        "type": "string",
        "description": "Unique ID for this document chunk"
    }},
    "content": {{
        "type": "string",
        "description": "A document chunk (subset of the original document)"
    }},
    "embeddings": {{
        "type": "vector",
        "description": "Vector embeddings for the document chunk."
    }},
    "metadata": {{
        "type": "jsonb",
        "description": "Metadata for the document chunk."
    }}
}}

{schema}

Rewrite the query so it works.

Output the final SQL query only.

SQL Query:
"""

DEFAULT_NUM_QUERY_RETRIES = 2


class LLMConfig(BaseModel):
    model_name: str = Field(
        default=DEFAULT_LLM_MODEL, description="LLM model to use for generation"
    )
    provider: str = Field(
        default=DEFAULT_LLM_MODEL_PROVIDER,
        description="LLM model provider to use for generation",
    )
    params: Dict[str, Any] = Field(default_factory=dict)


class MultiVectorRetrieverMode(Enum):
    """
    Enum for MultiVectorRetriever types.
    """

    SPLIT = "split"
    SUMMARIZE = "summarize"
    BOTH = "both"


class VectorStoreType(Enum):
    CHROMA = "chromadb"
    PGVECTOR = "pgvector"


vector_store_map = {VectorStoreType.CHROMA: Chroma, VectorStoreType.PGVECTOR: PGVector}


class VectorStoreConfig(BaseModel):
    vector_store_type: VectorStoreType = VectorStoreType.CHROMA
    persist_directory: str = None
    collection_name: str = DEFAULT_COLLECTION_NAME
    connection_string: str = None
    kb_table: Any = None
    is_sparse: bool = False
    vector_size: Optional[int] = None

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"


class RetrieverType(str, Enum):
    """Retriever type for RAG pipeline"""

    VECTOR_STORE = "vector_store"
    AUTO = "auto"
    MULTI = "multi"
    SQL = "sql"
    MULTI_HOP = "multi_hop"


class SearchType(Enum):
    """
    Enum for vector store search types.
    """

    SIMILARITY = "similarity"
    MMR = "mmr"
    SIMILARITY_SCORE_THRESHOLD = "similarity_score_threshold"


class SearchKwargs(BaseModel):
    k: int = Field(default=DEFAULT_K, description="Amount of documents to return", ge=1)
    filter: Optional[Dict[str, Any]] = Field(
        default=None, description="Filter by document metadata"
    )
    # For similarity_score_threshold search type
    score_threshold: Optional[float] = Field(
        default=None,
        description="Minimum relevance threshold for similarity_score_threshold search",
        ge=0.0,
        le=1.0,
    )
    # For MMR search type
    fetch_k: Optional[int] = Field(
        default=None, description="Amount of documents to pass to MMR algorithm", ge=1
    )
    lambda_mult: Optional[float] = Field(
        default=None,
        description="Diversity of results returned by MMR (1=min diversity, 0=max)",
        ge=0.0,
        le=1.0,
    )

    def model_dump(self, *args, **kwargs):
        # Override model_dump to exclude None values by default
        kwargs["exclude_none"] = True
        return super().model_dump(*args, **kwargs)


class LLMExample(BaseModel):
    input: str = Field(description="User input for the example")
    output: str = Field(
        description="What the LLM should generate for this example's input"
    )


class ValueSchema(BaseModel):
    value: Union[
        Union[str, int, float],
        Dict[Union[str, int, float], str],
        List[Union[str, int, float]],
    ] = Field(
        description="One of the following. The value as it exists in the table column. A dict of {table_value: descriptive value, ...}, where table_value is the value in the table. A list of sample values taken from the column."
    )
    comparator: Optional[Union[str, List[str]]] = Field(
        description="The posgtres sql operators used to compare two values. For example: `>`, `<`, `=`, or `%`.",
        default="=",
    )
    type: str = Field(
        description="A valid postgres type for this value. One of: int, string, float, or bool. When numbers appear they should be of type int or float."
    )
    description: str = Field(description="Description of what the value represents.")
    usage: str = Field(description="How and when to use this value for search.")
    example_questions: Optional[List[LLMExample]] = Field(
        default=None, description="Example questions where this value is set."
    )
    filter_threshold: Optional[float] = Field(
        default=0.0,
        description="Minimum relevance threshold to include metadata filters from this column.",
        exclude=True,
    )
    priority: Optional[int] = Field(
        default=0,
        description="Priority level for this column, lower numbers will be processed first.",
    )
    relevance: Optional[float] = Field(
        default=None,
        description="Relevance computed during search. Should not be set by the end user.",
        exclude=True,
    )


class MetadataConfig(BaseModel):
    """Class to configure metadata for retrieval. Only supports very basic document name lookup at the moment."""
    table: str = Field(
        description="Source table for metadata."
    )
    max_document_context: int = Field(
        # To work well with models with context window of 32768.
        default=16384,
        description="Truncate a document before using as context with an LLM if it exceeds this amount of tokens"
    )
    embeddings_table: str = Field(
        default="embeddings",
        description="Source table for embeddings"
    )
    id_column: str = Field(
        default="Id",
        description="Name of ID column in metadata table"
    )
    name_column: str = Field(
        default="Title",
        description="Name of column containing name or title of document"
    )
    name_column_index: Optional[str] = Field(
        default=None,
        description="Name of GIN index to use when looking up name."
    )
    content_column: str = Field(
        default="content",
        description="Name of column in embeddings table containing chunk content"
    )
    embeddings_metadata_column: str = Field(
        default="metadata",
        description="Name of column in embeddings table containing chunk metadata"
    )
    doc_id_key: str = Field(
        default="original_row_id",
        description="Metadata field that links an embedded chunk back to source document ID"
    )


class ColumnSchema(BaseModel):
    column: str = Field(description="Name of the column in the database")
    type: str = Field(description="Type of the column (e.g. int, string, datetime)")
    description: str = Field(description="Description of what the column represents")
    usage: str = Field(description="How and when to use this Table for search.")
    values: Optional[
        Union[
            OrderedDict[Union[str, int, float], ValueSchema],
            Dict[Union[str, int, float], ValueSchema],
        ]
    ] = Field(
        default=None,
        description="One of the following. A dict or ordered dict of {schema_value: ValueSchema, ...}, where schema value is the name given for this value description in the schema."
    )
    example_questions: Optional[List[LLMExample]] = Field(
        default=None, description="Example questions where this table is useful."
    )
    max_filters: Optional[int] = Field(
        default=1, description="Maximum number of filters to generate for this column."
    )
    filter_threshold: Optional[float] = Field(
        default=0.0,
        description="Minimum relevance threshold to include metadata filters from this column.",
    )
    priority: Optional[int] = Field(
        default=1,
        description="Priority level for this column, lower numbers will be processed first.",
    )
    relevance: Optional[float] = Field(
        default=None,
        description="Relevance computed during search. Should not be set by the end user.",
    )


class TableSchema(BaseModel):
    table: str = Field(description="Name of table in the database")
    description: str = Field(description="Description of what the table represents")
    usage: str = Field(description="How and when to use this Table for search.")
    columns: Optional[
        Union[OrderedDict[str, ColumnSchema], Dict[str, ColumnSchema]]
    ] = Field(
        description="Dict or Ordered Dict of {column_name: ColumnSchemas} describing the metadata columns available for the table"
    )
    example_questions: Optional[List[LLMExample]] = Field(
        default=None, description="Example questions where this table is useful."
    )
    join: str = Field(
        description="SQL join string to join this table with source documents table",
        default="",
    )
    max_filters: Optional[int] = Field(
        default=1, description="Maximum number of filters to generate for this table."
    )
    filter_threshold: Optional[float] = Field(
        default=0.0,
        description="Minimum relevance required to use this table to generate filters.",
    )
    priority: Optional[int] = Field(
        default=1,
        description="Priority level for this table, lower numbers will be processed first.",
    )
    relevance: Optional[float] = Field(
        default=None,
        description="Relevance computed during search. Should not be set by the end user.",
    )


class DatabaseSchema(BaseModel):
    database: str = Field(description="Name of database in the Database")
    description: str = Field(description="Description of what the Database represents")
    usage: str = Field(description="How and when to use this Database for search.")
    tables: Union[OrderedDict[str, TableSchema], Dict[str, TableSchema]] = Field(
        description="Dict of {column_name: ColumnSchemas} describing the metadata columns available for the table"
    )
    example_questions: Optional[List[LLMExample]] = Field(
        default=None, description="Example questions where this Database is useful."
    )
    max_filters: Optional[int] = Field(
        default=1,
        description="Maximum number of filters to generate for this Database.",
    )
    filter_threshold: Optional[float] = Field(
        default=0.0,
        description="Minimum relevance required to use this Database to generate filters.",
    )
    priority: Optional[int] = Field(
        default=0,
        description="Priority level for this Database, lower numbers will be processed first.",
    )
    relevance: Optional[float] = Field(
        default=None,
        description="Relevance computed during search. Should not be set by the end user.",
    )


class SQLRetrieverConfig(BaseModel):
    llm_config: LLMConfig = Field(
        default_factory=LLMConfig,
        description="LLM configuration to use for generating the final SQL query for retrieval",
    )
    metadata_filters_prompt_template: str = Field(
        default=DEFAULT_METADATA_FILTERS_PROMPT_TEMPLATE,
        description="Prompt template to generate PostgreSQL metadata filters. Has 'format_instructions', 'schema', 'examples', and 'input' input variables",
    )
    num_retries: int = Field(
        default=DEFAULT_NUM_QUERY_RETRIES,
        description="How many times for an LLM to try rewriting a failed SQL query before using the fallback retriever.",
    )
    rewrite_prompt_template: str = Field(
        default=DEFAULT_SEMANTIC_PROMPT_TEMPLATE,
        description="Prompt template to rewrite user input to be better suited for retrieval. Has 'input' input variable.",
    )
    table_prompt_template: str = Field(
        default=DEFAULT_TABLE_PROMPT_TEMPLATE,
        description="Prompt template to rewrite user input to be better suited for retrieval. Has 'input' input variable.",
    )
    column_prompt_template: str = Field(
        default=DEFAULT_COLUMN_PROMPT_TEMPLATE,
        description="Prompt template to rewrite user input to be better suited for retrieval. Has 'input' input variable.",
    )
    value_prompt_template: str = Field(
        default=DEFAULT_VALUE_PROMPT_TEMPLATE,
        description="Prompt template to rewrite user input to be better suited for retrieval. Has 'input' input variable.",
    )
    boolean_system_prompt: str = Field(
        default=DEFAULT_BOOLEAN_PROMPT_TEMPLATE,
        description="Prompt template to rewrite user input to be better suited for retrieval. Has 'input' input variable.",
    )
    generative_system_prompt: str = Field(
        default=DEFAULT_GENERATIVE_SYSTEM_PROMPT,
        description="Prompt template to rewrite user input to be better suited for retrieval. Has 'input' input variable.",
    )
    source_table: str = Field(
        description="Name of the source table containing the original documents that were embedded"
    )
    source_id_column: str = Field(
        description="Name of the column containing the UUID.", default="Id"
    )
    max_filters: Optional[int] = Field(
        description="Maximum number of filters to generate for sql queries.", default=10
    )
    filter_threshold: Optional[float] = Field(
        description="Minimum relevance required to use this Database to generate filters.",
        default=0.0,
    )
    min_k: Optional[int] = Field(
        description="Minimum number of documents accepted from a generated sql query.",
        default=10,
    )
    database_schema: Optional[DatabaseSchema] = Field(
        default=None,
        description="DatabaseSchema describing the database.",
    )
    examples: Optional[List[LLMExample]] = Field(
        default=None,
        description="Optional examples of final generated pgvector queries based on user input.",
    )


class SummarizationConfig(BaseModel):
    llm_config: LLMConfig = Field(
        default_factory=LLMConfig,
        description="LLM configuration to use for summarization",
    )
    map_prompt_template: str = Field(
        default=DEFAULT_MAP_PROMPT_TEMPLATE,
        description="Prompt for an LLM to summarize a single document",
    )
    reduce_prompt_template: str = Field(
        default=DEFAULT_REDUCE_PROMPT_TEMPLATE,
        description="Prompt for an LLM to summarize a set of summaries of documents into one",
    )
    max_summarization_tokens: int = Field(
        default=DEFAULT_MAX_SUMMARIZATION_TOKENS,
        description="Max number of tokens for summarized documents",
    )


class RerankerConfig(BaseModel):
    model: str = DEFAULT_RERANKING_MODEL
    base_url: str = DEFAULT_LLM_ENDPOINT
    filtering_threshold: float = 0.5
    num_docs_to_keep: Optional[int] = None
    max_concurrent_requests: int = 20
    max_retries: int = 3
    retry_delay: float = 1.0
    early_stop: bool = True  # Whether to enable early stopping
    early_stop_threshold: float = 0.8  # Confidence threshold for early stopping


class MultiHopRetrieverConfig(BaseModel):
    """Configuration for multi-hop retrieval"""

    base_retriever_type: RetrieverType = Field(
        default=RetrieverType.VECTOR_STORE,
        description="Type of base retriever to use for multi-hop retrieval",
    )
    max_hops: int = Field(
        default=3, description="Maximum number of follow-up questions to generate", ge=1
    )
    reformulation_template: str = Field(
        default=DEFAULT_QUESTION_REFORMULATION_TEMPLATE,
        description="Template for reformulating questions",
    )
    llm_config: LLMConfig = Field(
        default_factory=LLMConfig,
        description="LLM configuration to use for generating follow-up questions",
    )


class RAGPipelineModel(BaseModel):
    documents: Optional[List[Document]] = Field(
        default=None, description="List of documents"
    )

    vector_store_config: VectorStoreConfig = Field(
        default_factory=VectorStoreConfig, description="Vector store configuration"
    )

    llm: Optional[BaseChatModel] = Field(default=None, description="Language model")
    llm_model_name: str = Field(
        default=DEFAULT_LLM_MODEL, description="Language model name"
    )
    llm_provider: Optional[str] = Field(
        default=None, description="Language model provider"
    )
    vector_store: VectorStore = Field(
        default_factory=lambda: vector_store_map[VectorStoreConfig().vector_store_type],
        description="Vector store",
    )
    db_connection_string: Optional[str] = Field(
        default=None, description="Database connection string"
    )
    metadata_config: Optional[MetadataConfig] = Field(
        default=None,
        description="Configuration for metadata to be used for retrieval"
    )
    table_name: str = Field(default=DEFAULT_TEST_TABLE_NAME, description="Table name")
    embedding_model: Optional[Embeddings] = Field(
        default=None, description="Embedding model"
    )
    rag_prompt_template: str = Field(
        default=DEFAULT_RAG_PROMPT_TEMPLATE, description="RAG prompt template"
    )
    retriever_prompt_template: Optional[Union[str, dict]] = Field(
        default=None, description="Retriever prompt template"
    )
    retriever_type: RetrieverType = Field(
        default=RetrieverType.VECTOR_STORE, description="Retriever type"
    )
    search_type: SearchType = Field(
        default=SearchType.SIMILARITY, description="Type of search to perform"
    )
    search_kwargs: SearchKwargs = Field(
        default_factory=SearchKwargs,
        description="Search configuration for the retriever",
    )
    summarization_config: Optional[SummarizationConfig] = Field(
        default=None,
        description="Configuration for summarizing retrieved documents as context",
    )
    # SQL retriever specific.
    sql_retriever_config: Optional[SQLRetrieverConfig] = Field(
        default=None,
        description="Configuration for retrieving documents by generating SQL to filter by metadata & order by distance function",
    )

    # Multi retriever specific
    multi_retriever_mode: MultiVectorRetrieverMode = Field(
        default=MultiVectorRetrieverMode.BOTH, description="Multi retriever mode"
    )
    max_concurrency: int = Field(
        default=DEFAULT_MAX_CONCURRENCY, description="Maximum concurrency"
    )
    id_key: int = Field(default=DEFAULT_ID_KEY, description="ID key")
    parent_store: Optional[BaseStore] = Field(default=None, description="Parent store")
    text_splitter: Optional[TextSplitter] = Field(
        default=None, description="Text splitter"
    )
    chunk_size: int = Field(default=DEFAULT_CHUNK_SIZE, description="Chunk size")
    chunk_overlap: int = Field(
        default=DEFAULT_CHUNK_OVERLAP, description="Chunk overlap"
    )

    # Auto retriever specific
    auto_retriever_filter_columns: Optional[List[str]] = Field(
        default=None, description="Filter columns"
    )
    cardinality_threshold: int = Field(
        default=DEFAULT_CARDINALITY_THRESHOLD, description="Cardinality threshold"
    )
    content_column_name: str = Field(
        default=DEFAULT_CONTENT_COLUMN_NAME,
        description="Content column name (the column we will get embeddings)",
    )
    dataset_description: str = Field(
        default=DEFAULT_DATASET_DESCRIPTION, description="Description of the dataset"
    )
    reranker: bool = Field(
        default=DEFAULT_RERANKER_FLAG, description="Whether to use reranker"
    )
    reranker_config: RerankerConfig = Field(
        default_factory=RerankerConfig, description="Reranker configuration"
    )

    multi_hop_config: Optional[MultiHopRetrieverConfig] = Field(
        default=None,
        description="Configuration for multi-hop retrieval. Required when retriever_type is MULTI_HOP.",
    )

    @field_validator("multi_hop_config")
    @classmethod
    def validate_multi_hop_config(cls, v: Optional[MultiHopRetrieverConfig], info):
        """Validate that multi_hop_config is set when using multi-hop retrieval."""
        values = info.data
        if values.get("retriever_type") == RetrieverType.MULTI_HOP and v is None:
            raise ValueError(
                "multi_hop_config must be set when using multi-hop retrieval"
            )
        return v

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"

        json_schema_extra = {
            "example": {
                "retriever_type": RetrieverType.VECTOR_STORE.value,
                "multi_retriever_mode": MultiVectorRetrieverMode.BOTH.value,
                # add more examples here
            }
        }

    @classmethod
    def get_field_names(cls):
        return list(cls.model_fields.keys())

    @field_validator("search_kwargs")
    @classmethod
    def validate_search_kwargs(cls, v: SearchKwargs, info) -> SearchKwargs:
        search_type = info.data.get("search_type", SearchType.SIMILARITY)

        # Validate MMR-specific parameters
        if search_type == SearchType.MMR:
            if v.fetch_k is not None and v.fetch_k <= v.k:
                raise ValueError("fetch_k must be greater than k")
            if v.lambda_mult is not None and (v.lambda_mult < 0 or v.lambda_mult > 1):
                raise ValueError("lambda_mult must be between 0 and 1")
            if v.fetch_k is None and v.lambda_mult is not None:
                raise ValueError(
                    "fetch_k is required when using lambda_mult with MMR search type"
                )
            if v.lambda_mult is None and v.fetch_k is not None:
                raise ValueError(
                    "lambda_mult is required when using fetch_k with MMR search type"
                )
        elif search_type != SearchType.MMR:
            if v.fetch_k is not None:
                raise ValueError("fetch_k is only valid for MMR search type")
            if v.lambda_mult is not None:
                raise ValueError("lambda_mult is only valid for MMR search type")

        # Validate similarity_score_threshold parameters
        if search_type == SearchType.SIMILARITY_SCORE_THRESHOLD:
            if v.score_threshold is not None and (
                v.score_threshold < 0 or v.score_threshold > 1
            ):
                raise ValueError("score_threshold must be between 0 and 1")
            if v.score_threshold is None:
                raise ValueError(
                    "score_threshold is required for similarity_score_threshold search type"
                )
        elif (
            search_type != SearchType.SIMILARITY_SCORE_THRESHOLD
            and v.score_threshold is not None
        ):
            raise ValueError(
                "score_threshold is only valid for similarity_score_threshold search type"
            )

        return v



import os
import glob

from setuptools import find_packages, setup

# A special env var that allows us to disable the installation of the default extras for advanced users / containers / etc
MINDSDB_PIP_INSTALL_DEFAULT_EXTRAS = (
    True
    if os.getenv("MINDSDB_PIP_INSTALL_DEFAULT_EXTRAS", "true").lower() == "true"
    else False
)
DEFAULT_PIP_EXTRAS = [
    line.split("#")[0].rstrip()
    for line in open("default_handlers.txt").read().splitlines()
    if not line.strip().startswith("#")
]


class Deps:
    pkgs = []
    pkgs_exclude = ["tests", "tests.*"]
    new_links = []
    extras = {}


about = {}
with open("mindsdb/__about__.py") as fp:
    exec(fp.read(), about)


with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()


def expand_requirements_links(requirements: list) -> list:
    """Expand requirements that contain links to other requirement files"""
    to_add = []
    to_remove = []

    for requirement in requirements:
        if requirement.startswith("-r "):
            if os.path.exists(requirement.split()[1]):
                with open(requirement.split()[1]) as fh:
                    to_add += expand_requirements_links(
                        [req.strip() for req in fh.read().splitlines()]
                    )
            to_remove.append(requirement)

    for req in to_remove:
        requirements.remove(req)
    for req in to_add:
        requirements.append(req)

    return list(set(requirements))  # Remove duplicates


def define_deps():
    """Reads requirements.txt requirements-extra.txt files and preprocess it
    to be feed into setuptools.

    This is the only possible way (we found)
    how requirements.txt can be reused in setup.py
    using dependencies from private github repositories.

    Links must be appendend by `-{StringWithAtLeastOneNumber}`
    or something like that, so e.g. `-9231` works as well as
    `1.1.0`. This is ignored by the setuptools, but has to be there.

    Warnings:
        to make pip respect the links, you have to use
        `--process-dependency-links` switch. So e.g.:
        `pip install --process-dependency-links {git-url}`

    Returns:
         list of packages, extras and dependency links.
    """
    with open(os.path.normpath('requirements/requirements.txt')) as req_file:
        defaults = [req.strip() for req in req_file.read().splitlines()]

    links = []
    requirements = []
    for r in defaults:
        if 'git+https' in r:
            pkg = r.split('#')[-1]
            links.append(r + '-9876543210')
            requirements.append(pkg.replace('egg=', ''))
        else:
            requirements.append(r.strip())

    extra_requirements = {}
    full_requirements = []
    for fn in os.listdir(os.path.normpath('./requirements')):
        extra = []
        if fn.startswith('requirements-') and fn.endswith('.txt'):
            extra_name = fn.replace('requirements-', '').replace('.txt', '')
            with open(os.path.normpath(f"./requirements/{fn}")) as fp:
                extra = [req.strip() for req in fp.read().splitlines()]
            extra_requirements[extra_name] = extra
            full_requirements += extra

    extra_requirements['all_extras'] = list(set(full_requirements))

    full_handlers_requirements = []
    handlers_dir_path = os.path.normpath('./mindsdb/integrations/handlers')
    for fn in os.listdir(handlers_dir_path):
        if os.path.isdir(os.path.join(handlers_dir_path, fn)) and fn.endswith(
            "_handler"
        ):
            extra = []
            for req_file_path in glob.glob(
                os.path.join(handlers_dir_path, fn, "requirements*.txt")
            ):
                extra_name = fn.replace("_handler", "")
                file_name = os.path.basename(req_file_path)
                if file_name != "requirements.txt":
                    extra_name += "-" + file_name.replace("requirements_", "").replace(
                        ".txt", ""
                    )

                # If requirements.txt in our handler folder, import them as our extra's requirements
                if os.path.exists(req_file_path):
                    with open(req_file_path) as fp:
                        extra = expand_requirements_links(
                            [req.strip() for req in fp.read().splitlines()]
                        )

                    extra_requirements[extra_name] = extra
                    full_handlers_requirements += extra

                # Even with no requirements in our handler, list the handler as an extra (with no reqs)
                else:
                    extra_requirements[extra_name] = []

                # If this is a default extra and if we want to install defaults (enabled by default)
                #   then add it to the default requirements needing to install
                if (
                    MINDSDB_PIP_INSTALL_DEFAULT_EXTRAS
                    and extra_name in DEFAULT_PIP_EXTRAS
                    and extra
                ):
                    requirements += extra

    extra_requirements['all_handlers_extras'] = list(set(full_handlers_requirements))

    Deps.pkgs = requirements
    Deps.extras = extra_requirements
    Deps.new_links = links

    return Deps


deps = define_deps()

setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=deps.pkgs_exclude),
    install_requires=deps.pkgs,
    dependency_links=deps.new_links,
    extras_require=deps.extras,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10,<3.12",
)



"""data_integration_id

Revision ID: 96d5fef10caa
Revises: 473e8f239481
Create Date: 2022-09-06 15:20:02.382203

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db        # noqa


# revision identifiers, used by Alembic.
revision = '96d5fef10caa'
down_revision = '473e8f239481'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('data_integration_id', sa.Integer(), nullable=True))

    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    result = conn.execute(text('''
        select 1 from integration where name = 'lightwood';
    ''')).fetchall()
    if len(result) == 0:
        conn.execute(text('''
            insert into integration (name, engine, data)
            values ('lightwood', 'lightwood', '{}')
        '''))
    conn.execute(text('''
        update predictor set data_integration_id = integration_id
        where exists (select 1 from integration where integration.id = predictor.integration_id);
    '''))
    conn.execute(text('''
        update predictor set integration_id = (select id from integration where name = 'lightwood');
    '''))
    session.commit()

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'integration_id',
            existing_type=sa.INTEGER(),
            nullable=False
        )
        batch_op.create_foreign_key('fk_data_integration_id', 'integration', ['data_integration_id'], ['id'])


def downgrade():
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)
    conn.execute(text('''
        update predictor set integration_id = data_integration_id;
    '''))
    session.commit()

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_constraint('fk_data_integration_id', type_='foreignkey')
        batch_op.alter_column(
            'integration_id',
            existing_type=sa.INTEGER(),
            nullable=True
        )
        batch_op.drop_column('data_integration_id')



import json
import os
import shutil
from pathlib import Path

import pandas as pd

from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from sqlalchemy.orm.attributes import flag_modified

from mindsdb.integrations.utilities.files.file_reader import FileReader


logger = log.getLogger(__name__)


class FileController:
    def __init__(self):
        self.config = Config()
        self.fs_store = FsStore()
        self.dir = os.path.join(self.config.paths["content"], "file")

    def get_files_names(self):
        """return list of files names"""
        return [
            x[0]
            for x in db.session.query(db.File.name).filter_by(company_id=ctx.company_id)
        ]

    def get_file_meta(self, name):
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            return None
        columns = file_record.columns
        if isinstance(columns, str):
            columns = json.loads(columns)
        return {
            "name": file_record.name,
            "columns": columns,
            "row_count": file_record.row_count,
        }

    def get_files(self):
        """Get list of files

        Returns:
            list[dict]: files metadata
        """
        file_records = (
            db.session.query(db.File).filter_by(company_id=ctx.company_id).all()
        )
        files_metadata = [
            {
                "name": record.name,
                "row_count": record.row_count,
                "columns": record.columns,
            }
            for record in file_records
        ]
        return files_metadata

    def save_file(self, name, file_path, file_name=None):
        """Save the file to our store

        Args:
            name (str): with that name file will be available in sql api
            file_name (str): file name
            file_path (str): path to the file

        Returns:
            int: id of 'file' record in db
        """
        files_metadata = self.get_files()
        if name in [x["name"] for x in files_metadata]:
            raise Exception(f"File already exists: {name}")

        if file_name is None:
            file_name = Path(file_path).name

        file_dir = None
        try:
            pages_files, pages_index = self.get_file_pages(file_path)

            metadata = {
                'is_feather': True,
                'pages': pages_index
            }
            df = pages_files[0]
            file_record = db.File(
                name=name,
                company_id=ctx.company_id,
                source_file_path=file_name,
                file_path="",
                row_count=len(df),
                columns=list(df.columns),
                metadata_=metadata
            )
            db.session.add(file_record)
            db.session.flush()

            store_file_path = f"file_{ctx.company_id}_{file_record.id}"
            file_record.file_path = store_file_path

            file_dir = Path(self.dir).joinpath(store_file_path)
            file_dir.mkdir(parents=True, exist_ok=True)

            self.store_pages_as_feather(file_dir, pages_files)
            # store original file
            shutil.move(file_path, str(file_dir.joinpath(file_name)))

            self.fs_store.put(store_file_path, base_dir=self.dir)
            db.session.commit()

        except Exception as e:
            logger.error(e)
            if file_dir is not None:
                shutil.rmtree(file_dir)
            raise

        return file_record.id

    def get_file_pages(self, source_path: str):
        """
        Reads file and extract pages from it
        Returned structures:
          - page_files: dict with content, {page_num: dataframe}
          - pages_index: dict, link between page name and num: {page_name: page_num}
        """
        file_reader = FileReader(path=source_path)
        tables = file_reader.get_contents()

        pages_files = {}
        pages_index = {}
        if len(tables) == 1:
            df = list(tables.values())[0]
            pages_files[0] = df
        else:
            # file has several pages, create a new one with info
            df = pd.DataFrame(tables.keys(), columns=["Tables"])
            pages_files[0] = df
            for i, page_name in enumerate(tables.keys(), 1):
                pages_files[i] = tables[page_name]
                pages_index[page_name] = i
        return pages_files, pages_index

    def store_pages_as_feather(self, dest_dir: Path, pages_files: dict):
        """
        Stores pages in file storage dir in feather format
        """

        for num, df in pages_files.items():
            dest = dest_dir.joinpath(f'{num}.feather')
            df.to_feather(str(dest))

    def delete_file(self, name):
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            return None
        file_id = file_record.id
        db.session.delete(file_record)
        db.session.commit()
        self.fs_store.delete(f"file_{ctx.company_id}_{file_id}")
        return True

    def get_file_path(self, name):
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            raise Exception(f"File '{name}' does not exists")
        file_dir = f"file_{ctx.company_id}_{file_record.id}"
        self.fs_store.get(file_dir, base_dir=self.dir)
        return str(
            Path(self.dir)
            .joinpath(file_dir)
            .joinpath(Path(file_record.source_file_path).name)
        )

    def get_file_data(self, name: str, page_name: str = None) -> pd.DataFrame:
        """
        Returns file content as dataframe

        :param name: name of file
        :param page_name: page name, optional
        :return: Page or file content
        """
        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            raise Exception(f"File '{name}' does not exists")

        file_dir = f"file_{ctx.company_id}_{file_record.id}"
        self.fs_store.get(file_dir, base_dir=self.dir)

        metadata = file_record.metadata_ or {}
        if metadata.get('is_feather') is not True:
            # migrate file

            file_path = (
                Path(self.dir)
                .joinpath(file_dir)
                .joinpath(Path(file_record.source_file_path).name)
            )

            pages_files, pages_index = self.get_file_pages(str(file_path))

            self.store_pages_as_feather(file_path.parent, pages_files)
            metadata['is_feather'] = True
            metadata['pages'] = pages_index

            file_record.metadata_ = metadata
            flag_modified(file_record, 'metadata_')
            db.session.commit()

        if page_name is None:
            num = 0
        else:
            num = metadata.get('pages', {}).get(page_name)
            if num is None:
                raise KeyError(f'Page not found: {page_name}')

        path = (
            Path(self.dir)
            .joinpath(file_dir)
            .joinpath(f'{num}.feather')
        )
        return pd.read_feather(path)

    def set_file_data(self, name: str, df: pd.DataFrame, page_name: str = None):
        """
        Save file content
        :param name: name of file
        :param df: content to store
        :param page_name: name of page, optional
        """

        file_record = (
            db.session.query(db.File)
            .filter_by(company_id=ctx.company_id, name=name)
            .first()
        )
        if file_record is None:
            raise Exception(f"File '{name}' does not exists")

        file_dir = f"file_{ctx.company_id}_{file_record.id}"
        self.fs_store.get(file_dir, base_dir=self.dir)

        num = 0
        if page_name is not None and file_record.metadata_ is not None:
            num = file_record.metadata_.get('pages', {}).get(page_name, 0)

        path = (
            Path(self.dir)
            .joinpath(file_dir)
            .joinpath(f'{num}.feather')
        )
        df.to_feather(path)
        self.fs_store.put(file_dir, base_dir=self.dir)



"""add_llm_log_columns

Revision ID: 9d559f68d535
Revises: 6c57ed39a82b
Create Date: 2024-11-15 11:24:28.808881

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '9d559f68d535'
down_revision = '6c57ed39a82b'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('llm_log', schema=None) as batch_op:
        batch_op.alter_column('company_id', nullable=False)
        batch_op.alter_column('model_id', nullable=True)
        batch_op.add_column(sa.Column('model_group', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('cost', sa.Numeric(), nullable=True))
        batch_op.add_column(sa.Column('exception', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('traceback', sa.String(), nullable=True))
        batch_op.add_column(sa.Column('stream', sa.Boolean(), default=False))
        batch_op.add_column(sa.Column('metadata', sa.JSON(), nullable=True))


def downgrade():
    with op.batch_alter_table('llm_log', schema=None) as batch_op:
        batch_op.alter_column('company_id', nullable=True)
        batch_op.alter_column('model_id', nullable=False)
        batch_op.drop_column('model_group')
        batch_op.drop_column('cost')
        batch_op.drop_column('exception')
        batch_op.drop_column('traceback')
        batch_op.drop_column('stream')
        batch_op.drop_column('metadata')



import unittest

from newsapi.newsapi_exception import NewsAPIException

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.handlers.newsapi_handler.newsapi_handler import NewsAPIHandler


class NewsApiHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {"api_key": "82fa480335ce42c0aa3758cb0efe66be"}
        }
        cls.handler = NewsAPIHandler("test_newsapi_handler", **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        self.handler.check_connection()

    def test_2_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query('SELECT * FROM article WHERE query="google"')
        assert res.type is RESPONSE_TYPE.TABLE

    def test_3_select(self):
        # table = self.handler.get_table("article")
        with self.assertRaises(NewsAPIException):
            res = self.handler.native_query("SELECT * FROM article")

    def test_4_select(self):
        # table = self.handler.get_table("article")
        with self.assertRaises(NewsAPIException):
            res = self.handler.native_query(
                'SELECT * FROM article WHERE query="google" AND sources="google.com"'
            )

    def test_5_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query(
            'SELECT * FROM article WHERE query="google" AND sources="abc-news"'
        )
        assert res.type is RESPONSE_TYPE.TABLE

    def test_6_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query(
            'SELECT * FROM article WHERE query="google" AND publishedAt >= "2023-03-23" AND  publishedAt <= "2023-04-23"'
        )

    def test_7_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query(
            'SELECT * FROM article WHERE query="google" LIMIT 78'
        )
        assert res.type is RESPONSE_TYPE.TABLE

    def test_8_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query(
            'SELECT * FROM article WHERE query="google" LIMIT 150'
        )
        assert res.type is RESPONSE_TYPE.TABLE

    def test_9_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query(
            'SELECT * FROM article WHERE query="google" ORDER BY publishedAt'
        )
        assert res.type is RESPONSE_TYPE.TABLE

    def test_10_select(self):
        # table = self.handler.get_table("article")
        with self.assertRaises(NotImplementedError):
            res = self.handler.native_query(
                'SELECT * FROM article WHERE query="google" ORDER BY query'
            )

    def test_11_select(self):
        # table = self.handler.get_table("article")
        res = self.handler.native_query(
            'SELECT * FROM article WHERE query="google" ORDER BY relevancy'
        )
        assert res.type is RESPONSE_TYPE.TABLE


if __name__ == "__main__":
    unittest.main()



from concurrent.futures import as_completed, TimeoutError
from typing import Optional, Dict
import re

from langchain.agents import AgentExecutor
from langchain.agents.initialize import initialize_agent
from langchain.chains.conversation.memory import ConversationSummaryBufferMemory
from langchain_core.prompts import PromptTemplate
from langchain_core.messages import SystemMessage, HumanMessage

import numpy as np
import pandas as pd

from mindsdb.interfaces.agents.safe_output_parser import SafeOutputParser
from mindsdb.interfaces.agents.langchain_agent import (
    get_llm_provider, get_embedding_model_provider, create_chat_model, get_chat_model_params
)

from mindsdb.interfaces.agents.constants import (
    DEFAULT_AGENT_TIMEOUT_SECONDS,
    DEFAULT_AGENT_TOOLS,
    DEFAULT_AGENT_TYPE,
    DEFAULT_MAX_ITERATIONS,
    DEFAULT_MAX_TOKENS,
    DEFAULT_MODEL_NAME,
    USER_COLUMN,
    ASSISTANT_COLUMN
)
from mindsdb.integrations.utilities.rag.settings import DEFAULT_RAG_PROMPT_TEMPLATE
from mindsdb.integrations.handlers.langchain_handler.tools import setup_tools
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.interfaces.storage.model_fs import HandlerStorage, ModelStorage
from mindsdb.integrations.handlers.langchain_embedding_handler.langchain_embedding_handler import construct_model_from_args
from mindsdb.integrations.handlers.openai_handler.constants import CHAT_MODELS  # noqa, for dependency checker

from mindsdb.utilities import log
from mindsdb.utilities.context_executor import ContextThreadPoolExecutor

_PARSING_ERROR_PREFIXES = ['An output parsing error occured', 'Could not parse LLM output']

logger = log.getLogger(__name__)


class LangChainHandler(BaseMLEngine):
    """
    This is a MindsDB integration for the LangChain library, which provides a unified interface for interacting with
    various large language models (LLMs).

    Supported LLM providers:
        - OpenAI
        - Anthropic
        - Anyscale
        - Google
        - LiteLLM
        - Ollama

    Supported standard tools:
        - python_repl
        - serper.dev search
    """
    name = 'langchain'

    def __init__(
            self,
            model_storage: ModelStorage,
            engine_storage: HandlerStorage,
            **kwargs):
        super().__init__(model_storage, engine_storage, **kwargs)
        # if True, the target column name does not have to be specified at creation time.
        self.generative = True
        self.default_agent_tools = DEFAULT_AGENT_TOOLS

    def _create_embeddings_model(self, args: Dict):
        return construct_model_from_args(args)

    def _handle_parsing_errors(self, error: Exception) -> str:
        response = str(error)
        for p in _PARSING_ERROR_PREFIXES:
            if response.startswith(p):
                # As a somewhat dirty workaround, we accept the output formatted incorrectly and use it as a response.
                #
                # Ideally, in the future, we would write a parser that is more robust and flexible than the one Langchain uses.
                # Response is wrapped in ``
                logger.info('Handling parsing error, salvaging response...')
                response_output = response.split('`')
                if len(response_output) >= 2:
                    response = response_output[-2]

                # Wrap response in Langchain conversational react format.
                langchain_react_formatted_response = f'''Thought: Do I need to use a tool? No
AI: {response}'''
                return langchain_react_formatted_response
        return f'Agent failed with error:\n{str(error)}...'

    def create(self, target: str, args: Dict = None, **kwargs):
        self.default_agent_tools = args.get('tools', self.default_agent_tools)

        args = args['using']
        args['target'] = target
        args['model_name'] = args.get('model_name', DEFAULT_MODEL_NAME)
        args['provider'] = args.get('provider', get_llm_provider(args))
        args['embedding_model_provider'] = args.get('embedding_model', get_embedding_model_provider(args))
        if args.get('mode') == 'retrieval':
            # use default prompt template for retrieval i.e. RAG if not provided
            if "prompt_template" not in args:
                args["prompt_template"] = DEFAULT_RAG_PROMPT_TEMPLATE

        self.model_storage.json_set('args', args)

    @staticmethod
    def create_validation(_, args: Dict=None, **kwargs):
        if 'using' not in args:
            raise Exception("LangChain engine requires a USING clause! Refer to its documentation for more details.")
        else:
            args = args['using']
        if 'prompt_template' not in args:
            if not args.get('mode') == 'retrieval':
                raise ValueError('Please provide a `prompt_template` for this engine.')

    def predict(self, df: pd.DataFrame, args: Dict=None) -> pd.DataFrame:
        """
        Dispatch is performed depending on the underlying model type. Currently, only the default text completion
        is supported.
        """
        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get('args')
        if 'prompt_template' not in args and 'prompt_template' not in pred_args:
            raise ValueError(f"This model expects a `prompt_template`, please provide one.")
        # Back compatibility for old models
        args['provider'] = args.get('provider', get_llm_provider(args))
        args['embedding_model_provider'] = args.get('embedding_model', get_embedding_model_provider(args))

        df = df.reset_index(drop=True)

        if pred_args.get('mode') == 'chat_model':
            return self.call_llm(df, args, pred_args)

        agent = self.create_agent(df, args, pred_args)
        # Use last message as prompt, remove other questions.
        user_column = args.get('user_column', USER_COLUMN)
        if user_column not in df.columns:
            raise Exception(
                f"Expected user input in column `{user_column}`, which is not found in the input data. Either provide the column, or redefine the expected column at model creation (`USING user_column = 'value'`)")  # noqa
        df.iloc[:-1, df.columns.get_loc(user_column)] = None
        return self.run_agent(df, agent, args, pred_args)

    def call_llm(self, df, args=None, pred_args=None):
        llm = create_chat_model({**args, **pred_args})

        user_column = args.get('user_column', USER_COLUMN)
        assistant_column = args.get('assistant_column', ASSISTANT_COLUMN)

        question = df[user_column].iloc[-1]
        resp = llm([HumanMessage(question)], stop=['\nObservation:', '\n\tObservation:'])

        return pd.DataFrame([resp.content], columns=[assistant_column])

    def create_agent(self, df: pd.DataFrame, args: Dict=None, pred_args: Dict=None) -> AgentExecutor:
        pred_args = pred_args if pred_args else {}

        # Set up tools.
        model_kwargs = get_chat_model_params({**args, **pred_args})
        llm = create_chat_model({**args, **pred_args})

        tools = setup_tools(llm,
                            model_kwargs,
                            pred_args,
                            self.default_agent_tools)

        # Prefer prediction prompt template over original if provided.
        prompt_template = pred_args.get('prompt_template', args['prompt_template'])
        if 'context' in pred_args:
            prompt_template += '\n\n' + 'Useful information:\n' + pred_args['context'] + '\n'

        # Set up memory.
        memory = ConversationSummaryBufferMemory(llm=llm,
                                                 max_token_limit=model_kwargs.get('max_tokens', DEFAULT_MAX_TOKENS),
                                                 memory_key='chat_history')
        memory.chat_memory.messages.insert(0, SystemMessage(content=prompt_template))
        # User - Assistant conversation. All except the last message.
        user_column = args.get('user_column', USER_COLUMN)
        assistant_column = args.get('assistant_column', ASSISTANT_COLUMN)
        for row in df[:-1].to_dict('records'):
            question = row[user_column]
            answer = row[assistant_column]
            if question:
                memory.chat_memory.add_user_message(question)
            if answer:
                memory.chat_memory.add_ai_message(answer)

        agent_type = args.get('agent_type', DEFAULT_AGENT_TYPE)
        agent_executor = initialize_agent(
            tools,
            llm,
            agent=agent_type,
            # Use custom output parser to handle flaky LLMs that don't ALWAYS conform to output format.
            agent_kwargs={'output_parser': SafeOutputParser()},
            # Calls the agent’s LLM Chain one final time to generate a final answer based on the previous steps
            early_stopping_method='generate',
            handle_parsing_errors=self._handle_parsing_errors,
            # Timeout per agent invocation.
            max_execution_time=pred_args.get('timeout_seconds', args.get('timeout_seconds', DEFAULT_AGENT_TIMEOUT_SECONDS)),
            max_iterations=pred_args.get('max_iterations', args.get('max_iterations', DEFAULT_MAX_ITERATIONS)),
            memory=memory,
            verbose=pred_args.get('verbose', args.get('verbose', True))
        )
        return agent_executor

    def run_agent(self, df: pd.DataFrame, agent: AgentExecutor, args: Dict, pred_args: Dict) -> pd.DataFrame:
        # Prefer prediction time prompt template, if available.
        base_template = pred_args.get('prompt_template', args['prompt_template'])

        input_variables = []
        matches = list(re.finditer("{{(.*?)}}", base_template))

        for m in matches:
            input_variables.append(m[0].replace('{', '').replace('}', ''))
        empty_prompt_ids = np.where(df[input_variables].isna().all(axis=1).values)[0]

        base_template = base_template.replace('{{', '{').replace('}}', '}')
        prompts = []

        user_column = args.get('user_column', USER_COLUMN)
        for i, row in df.iterrows():
            if i not in empty_prompt_ids:
                prompt = PromptTemplate(input_variables=input_variables, template=base_template)
                kwargs = {}
                for col in input_variables:
                    kwargs[col] = row[col] if row[col] is not None else ''  # add empty quote if data is missing
                prompts.append(prompt.format(**kwargs))
            elif row.get(user_column):
                # Just add prompt
                prompts.append(row[user_column])

        def _invoke_agent_executor_with_prompt(agent_executor, prompt):
            if not prompt:
                return ''
            try:
                answer = agent_executor.invoke(prompt)
            except Exception as e:
                answer = str(e)
                if not answer.startswith("Could not parse LLM output: `"):
                    raise e
                answer = {'output': answer.removeprefix("Could not parse LLM output: `").removesuffix("`")}

            if 'output' not in answer:
                # This should never happen unless Langchain changes invoke output format, but just in case.
                return agent_executor.run(prompt)
            return answer['output']

        completions = []
        # max_workers defaults to number of processors on the machine multiplied by 5.
        # https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
        max_workers = args.get('max_workers', None)
        agent_timeout_seconds = args.get('timeout', DEFAULT_AGENT_TIMEOUT_SECONDS)
        executor = ContextThreadPoolExecutor(max_workers=max_workers)
        futures = [executor.submit(_invoke_agent_executor_with_prompt, agent, prompt) for prompt in prompts]
        try:
            for future in as_completed(futures, timeout=agent_timeout_seconds):
                completions.append(future.result())
        except TimeoutError:
            completions.append("I'm sorry! I couldn't come up with a response in time. Please try again.")
        # Can't use ThreadPoolExecutor as context manager since we need wait=False.
        executor.shutdown(wait=False)

        # Add null completion for empty prompts
        for i in sorted(empty_prompt_ids)[:-1]:
            completions.insert(i, None)

        pred_df = pd.DataFrame(completions, columns=[args['target']])

        return pred_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        tables = ['info']
        return pd.DataFrame(tables, columns=['tables'])

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        raise NotImplementedError('Fine-tuning is not supported for LangChain models')



import sys
import copy
import datetime as dt
from copy import deepcopy
from multiprocessing.pool import ThreadPool

import pandas as pd
from dateutil.parser import parse as parse_datetime

from sqlalchemy import func
import numpy as np

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.config import Config
from mindsdb.interfaces.model.functions import (
    get_model_record,
    get_model_records,
    get_project_record
)
from mindsdb.interfaces.storage.json import get_json_storage
from mindsdb.interfaces.storage.model_fs import ModelStorage
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.functions import resolve_model_identifier
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError
from mindsdb.utilities import log

logger = log.getLogger(__name__)

IS_PY36 = sys.version_info[1] <= 6

default_project = config.get('default_project')


def delete_model_storage(model_id, ctx_dump):
    try:
        ctx.load(ctx_dump)
        modelStorage = ModelStorage(model_id)
        modelStorage.delete()
    except Exception as e:
        logger.error(f'Something went wrong during deleting storage of model {model_id}: {e}')


class ModelController():
    config: Config

    def __init__(self) -> None:
        self.config = Config()

    def get_model_data(self, name: str = None, predictor_record=None, ml_handler_name='lightwood') -> dict:
        if predictor_record is None:
            predictor_record = get_model_record(except_absent=True, name=name, ml_handler_name=ml_handler_name)

        data = deepcopy(predictor_record.data)
        data['dtype_dict'] = predictor_record.dtype_dict
        data['created_at'] = str(parse_datetime(str(predictor_record.created_at).split('.')[0]))
        data['updated_at'] = str(parse_datetime(str(predictor_record.updated_at).split('.')[0]))
        data['training_start_at'] = predictor_record.training_start_at
        data['training_stop_at'] = predictor_record.training_stop_at
        data['predict'] = predictor_record.to_predict[0]
        data['update'] = predictor_record.update_status
        data['mindsdb_version'] = predictor_record.mindsdb_version
        data['name'] = predictor_record.name
        data['code'] = predictor_record.code
        data['problem_definition'] = predictor_record.learn_args
        data['fetch_data_query'] = predictor_record.fetch_data_query
        data['active'] = predictor_record.active
        data['status'] = predictor_record.status
        data['id'] = predictor_record.id
        data['version'] = predictor_record.version

        json_storage = get_json_storage(
            resource_id=predictor_record.id
        )
        data['json_ai'] = json_storage.get('json_ai')

        if data.get('accuracies', None) is not None:
            if len(data['accuracies']) > 0:
                data['accuracy'] = float(np.mean(list(data['accuracies'].values())))
        return data

    def get_reduced_model_data(self, name: str = None, predictor_record=None, ml_handler_name='lightwood') -> dict:
        full_model_data = self.get_model_data(name=name, predictor_record=predictor_record, ml_handler_name=ml_handler_name)
        reduced_model_data = {}
        for k in ['id', 'name', 'version', 'is_active', 'predict', 'status',
                  'problem_definition', 'current_phase', 'accuracy', 'data_source', 'update', 'active',
                  'mindsdb_version', 'error', 'created_at', 'fetch_data_query']:
            reduced_model_data[k] = full_model_data.get(k, None)

        reduced_model_data['training_time'] = None
        if full_model_data.get('training_start_at') is not None:
            if full_model_data.get('training_stop_at') is not None:
                reduced_model_data['training_time'] = (
                    full_model_data.get('training_stop_at')
                    - full_model_data.get('training_start_at')
                )
            elif full_model_data.get('status') == 'training':
                reduced_model_data['training_time'] = (
                    dt.datetime.now()
                    - full_model_data.get('training_start_at')
                )
            if reduced_model_data['training_time'] is not None:
                reduced_model_data['training_time'] = (
                    reduced_model_data['training_time']
                    - dt.timedelta(microseconds=reduced_model_data['training_time'].microseconds)
                ).total_seconds()

        return reduced_model_data

    def describe_model(self, session, project_name, model_name, attribute, version=None):
        args = {
            'name': model_name,
            'version': version,
            'project_name': project_name,
            'except_absent': True
        }
        if version is not None:
            args['active'] = None

        model_record = get_model_record(**args)

        integration_record = db.Integration.query.get(model_record.integration_id)

        ml_handler_base = session.integration_controller.get_ml_handler(integration_record.name)

        return ml_handler_base.describe(model_record.id, attribute)

    def get_model(self, name, version=None, ml_handler_name=None, project_name=None):
        show_active = True if version is None else None
        model_record = get_model_record(
            active=show_active,
            version=version,
            name=name,
            ml_handler_name=ml_handler_name,
            except_absent=True,
            project_name=project_name)
        data = self.get_reduced_model_data(predictor_record=model_record)
        integration_record = db.Integration.query.get(model_record.integration_id)
        if integration_record is not None:
            data['engine'] = integration_record.engine
            data['engine_name'] = integration_record.name
        return data

    def get_models(self, with_versions=False, ml_handler_name=None, integration_id=None,
                   project_name=None):
        models = []
        show_active = True if with_versions is False else None
        for model_record in get_model_records(active=show_active, ml_handler_name=ml_handler_name,
                                              integration_id=integration_id, project_name=project_name):
            model_data = self.get_reduced_model_data(predictor_record=model_record)
            models.append(model_data)
        return models

    def delete_model(self, model_name: str, project_name: str = default_project, version=None):
        from mindsdb.interfaces.database.database import DatabaseController

        project_record = get_project_record(func.lower(project_name))
        if project_record is None:
            raise Exception(f"Project '{project_name}' does not exists")

        database_controller = DatabaseController()

        project = database_controller.get_project(project_name)

        if version is None:
            # Delete latest version
            predictors_records = get_model_records(
                name=model_name,
                project_id=project.id,
                active=None,
            )
        else:
            predictors_records = get_model_records(
                name=model_name,
                project_id=project.id,
                version=version,
            )
        if len(predictors_records) == 0:
            raise EntityNotExistsError('Model does not exist', model_name)

        is_cloud = self.config.get('cloud', False)
        if is_cloud:
            for predictor_record in predictors_records:
                model_data = self.get_model_data(predictor_record=predictor_record)
                if (
                    model_data.get('status') in ['generating', 'training']
                    and isinstance(model_data.get('created_at'), str) is True
                    and (dt.datetime.now() - parse_datetime(model_data.get('created_at'))) < dt.timedelta(hours=1)
                ):
                    raise Exception('You are unable to delete models currently in progress, please wait before trying again')

        for predictor_record in predictors_records:
            if is_cloud:
                predictor_record.deleted_at = dt.datetime.now()
            else:
                db.session.delete(predictor_record)
        db.session.commit()

        # region delete storages
        if len(predictors_records) > 1:
            ctx_dump = ctx.dump()
            with ThreadPool(min(len(predictors_records), 100)) as pool:
                pool.starmap(delete_model_storage, [(record.id, ctx_dump) for record in predictors_records])
        else:
            modelStorage = ModelStorage(predictors_records[0].id)
            modelStorage.delete()
        # endregion

    def rename_model(self, old_name, new_name):
        model_record = get_model_record(name=new_name)
        if model_record is None:
            raise Exception(f"Model with name '{new_name}' already exists")

        for model_record in get_model_records(name=old_name):
            model_record.name = new_name
        db.session.commit()

    @staticmethod
    def _get_data_integration_ref(statement, database_controller):
        # TODO use database_controller handler_controller internally
        data_integration_ref = None
        fetch_data_query = None
        if statement.integration_name is not None:
            fetch_data_query = statement.query_str
            integration_name = statement.integration_name.parts[0].lower()

            databases_meta = database_controller.get_dict()
            if integration_name not in databases_meta:
                raise EntityNotExistsError('Database does not exist', integration_name)
            data_integration_meta = databases_meta[integration_name]
            # TODO improve here. Suppose that it is view
            if data_integration_meta['type'] == 'project':
                data_integration_ref = {
                    'type': 'project'
                }
            elif data_integration_meta['type'] == 'system':
                data_integration_ref = {
                    'type': 'system'
                }
            else:
                data_integration_ref = {
                    'type': 'integration',
                    'id': data_integration_meta['id']
                }
        return data_integration_ref, fetch_data_query

    def prepare_create_statement(self, statement, database_controller):
        # extract data from Create model or Retrain statement and prepare it for using in crate and retrain functions
        project_name = statement.name.parts[0].lower()
        model_name = statement.name.parts[1].lower()

        sql_task = None
        if statement.task is not None:
            sql_task = statement.task.to_string()
        problem_definition = {
            '__mdb_sql_task': sql_task
        }
        if statement.targets is not None:
            problem_definition['target'] = statement.targets[0].parts[-1]

        data_integration_ref, fetch_data_query = self._get_data_integration_ref(statement, database_controller)

        label = None
        if statement.using is not None:
            label = statement.using.pop('tag', None)

            problem_definition['using'] = statement.using

        if statement.order_by is not None:
            problem_definition['timeseries_settings'] = {
                'is_timeseries': True,
                'order_by': getattr(statement, 'order_by')[0].field.parts[-1]
            }
            for attr in ['horizon', 'window']:
                if getattr(statement, attr) is not None:
                    problem_definition['timeseries_settings'][attr] = getattr(statement, attr)

            if statement.group_by is not None:
                problem_definition['timeseries_settings']['group_by'] = [col.parts[-1] for col in statement.group_by]

        join_learn_process = False
        if 'join_learn_process' in problem_definition.get('using', {}):
            join_learn_process = problem_definition['using']['join_learn_process']
            del problem_definition['using']['join_learn_process']

        return dict(
            model_name=model_name,
            project_name=project_name,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            problem_definition=problem_definition,
            join_learn_process=join_learn_process,
            label=label
        )

    def create_model(self, statement, ml_handler):
        params = self.prepare_create_statement(statement, ml_handler.database_controller)

        existing_projects_meta = ml_handler.database_controller.get_dict(filter_type='project')
        if params['project_name'] not in existing_projects_meta:
            raise EntityNotExistsError('Project does not exist', params['project_name'])

        project = ml_handler.database_controller.get_project(name=params['project_name'])
        project_tables = project.get_tables()
        if params['model_name'] in project_tables:
            raise EntityExistsError('Model already exists', f"{params['project_name']}.{params['model_name']}")
        predictor_record = ml_handler.learn(**params)

        return ModelController.get_model_info(predictor_record)

    def retrain_model(self, statement, ml_handler):
        # active setting
        set_active = True
        if statement.using is not None:
            set_active = statement.using.pop('active', True)
            if set_active in ('0', 0, None):
                set_active = False

        params = self.prepare_create_statement(statement, ml_handler.database_controller)

        base_predictor_record = get_model_record(
            name=params['model_name'],
            project_name=params['project_name'],
            active=True
        )

        model_name = params['model_name']
        if base_predictor_record is None:
            raise Exception(f"Error: model '{model_name}' does not exist")

        if params['data_integration_ref'] is None:
            params['data_integration_ref'] = base_predictor_record.data_integration_ref
        if params['fetch_data_query'] is None:
            params['fetch_data_query'] = base_predictor_record.fetch_data_query

        problem_definition = base_predictor_record.learn_args.copy()
        problem_definition.update(params['problem_definition'])
        params['problem_definition'] = problem_definition

        params['is_retrain'] = True
        params['set_active'] = set_active
        predictor_record = ml_handler.learn(**params)

        return ModelController.get_model_info(predictor_record)

    def prepare_finetune_statement(self, statement, database_controller):
        project_name, model_name, model_version = resolve_model_identifier(statement.name)
        if project_name is None:
            project_name = default_project
        data_integration_ref, fetch_data_query = self._get_data_integration_ref(statement, database_controller)

        set_active = True
        if statement.using is not None:
            set_active = statement.using.pop('active', True)
            if set_active in ('0', 0, None):
                set_active = False

        label = None
        args = {}
        if statement.using is not None:
            label = statement.using.pop('tag', None)
            args = statement.using

        join_learn_process = args.pop('join_learn_process', False)

        base_predictor_record = get_model_record(
            name=model_name,
            project_name=project_name,
            version=model_version,
            active=True if model_version is None else None
        )

        if data_integration_ref is None:
            data_integration_ref = base_predictor_record.data_integration_ref
        if fetch_data_query is None:
            fetch_data_query = base_predictor_record.fetch_data_query

        return dict(
            model_name=model_name,
            project_name=project_name,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            base_model_version=model_version,
            args=args,
            join_learn_process=join_learn_process,
            label=label,
            set_active=set_active
        )

    @profiler.profile()
    def finetune_model(self, statement, ml_handler):
        params = self.prepare_finetune_statement(statement, ml_handler.database_controller)
        predictor_record = ml_handler.finetune(**params)
        return ModelController.get_model_info(predictor_record)

    def update_model(self, session, project_name: str, model_name: str, problem_definition, version=None):

        model_record = get_model_record(
            name=model_name,
            version=version,
            project_name=project_name,
            except_absent=True
        )
        integration_record = db.Integration.query.get(model_record.integration_id)

        ml_handler_base = session.integration_controller.get_ml_handler(integration_record.name)
        ml_handler_base.update(args=problem_definition, model_id=model_record.id)

        # update model record
        if 'using' in problem_definition:
            learn_args = copy.deepcopy(model_record.learn_args)
            learn_args['using'].update(problem_definition['using'])
            model_record.learn_args = learn_args
            db.session.commit()

    @staticmethod
    def get_model_info(predictor_record):
        from mindsdb.interfaces.database.projects import ProjectController
        projects_controller = ProjectController()
        project = projects_controller.get(id=predictor_record.project_id)

        columns = ['NAME', 'ENGINE', 'PROJECT', 'ACTIVE', 'VERSION', 'STATUS', 'ACCURACY', 'PREDICT', 'UPDATE_STATUS',
                   'MINDSDB_VERSION', 'ERROR', 'SELECT_DATA_QUERY', 'TRAINING_OPTIONS', 'TAG']

        project_name = project.name
        model = project.get_model_by_id(model_id=predictor_record.id)
        table_name = model['name']
        table_meta = model['metadata']
        record = [
            table_name, table_meta['engine'], project_name, table_meta['active'], table_meta['version'], table_meta['status'],
            table_meta['accuracy'], table_meta['predict'], table_meta['update_status'],
            table_meta['mindsdb_version'], table_meta['error'], table_meta['select_data_query'],
            str(table_meta['training_options']), table_meta['label']
        ]

        return pd.DataFrame([record], columns=columns)

    def set_model_active_version(self, project_name, model_name, version):

        model_record = get_model_record(
            name=model_name,
            project_name=project_name,
            version=version,
            active=None
        )

        if model_record is None:
            raise EntityNotExistsError(f'Model {model_name} with version {version} is not found in {project_name}')

        model_record.active = True

        # deactivate current active version
        model_records = db.Predictor.query.filter(
            db.Predictor.name == model_record.name,
            db.Predictor.project_id == model_record.project_id,
            db.Predictor.active == True,    # noqa
            db.Predictor.company_id == ctx.company_id,
            db.Predictor.id != model_record.id
        )
        for p in model_records:
            p.active = False

        db.session.commit()

    def delete_model_version(self, project_name, model_name, version):

        model_record = get_model_record(
            name=model_name,
            project_name=project_name,
            version=version,
            active=None
        )
        if model_record is None:
            raise EntityNotExistsError(f'Model {model_name} with version {version} is not found in {project_name}')

        if model_record.active:
            raise Exception(f"Can't remove active version: {project_name}.{model_name}.{version}")

        is_cloud = self.config.get('cloud', False)
        if is_cloud:
            model_record.deleted_at = dt.datetime.now()
        else:
            db.session.delete(model_record)
        modelStorage = ModelStorage(model_record.id)
        modelStorage.delete()

        db.session.commit()



from pathlib import Path

from alembic.command import upgrade, autogen    # noqa
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.script.revision import ResolutionError
from alembic.migration import MigrationContext

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def migrate_to_head():
    """ Trying to update database to head revision.
        If alembic unable to recognize current revision (In case when database version is newer than backend)
        then do nothing.
    """

    config_file = Path(__file__).parent / 'alembic.ini'
    config = Config(config_file)

    # mindsdb can runs not from project directory
    script_location_abc = config_file.parent / config.get_main_option('script_location')
    config.set_main_option('script_location', str(script_location_abc))

    script = ScriptDirectory.from_config(config)
    with db.engine.begin() as conn:
        mc = MigrationContext.configure(conn)
        cur_revision = mc.get_current_revision()

        try:
            script.revision_map.get_revision(cur_revision)
        except ResolutionError:
            raise Exception("Database version higher than application.")

    head_rev = script.get_current_head()
    if cur_revision == head_rev:
        logger.debug("The database is in its current state, no updates are required.")
        return

    logger.info("Migrations are available. Applying updates to the database.")
    upgrade(config=config, revision='head')


if __name__ == "__main__":
    # have to import this because
    # all env initialization happens here
    from mindsdb.utilities.config import Config as MDBConfig
    MDBConfig()
    db.init()
    migrate_to_head()



from mindsdb_sql_parser.ast import (
    Identifier,
    Constant,
    Select,
    Join,
    Parameter,
    BinaryOperation,
    Tuple,
)
from mindsdb.api.executor.planner.steps import FetchDataframeStep
from mindsdb.integrations.utilities.query_traversal import query_traversal

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.api.executor.exceptions import UnknownError
from mindsdb.interfaces.query_context.context_controller import query_context_controller

from .base import BaseStepCall


def get_table_alias(table_obj, default_db_name):
    # (database, table, alias)
    if isinstance(table_obj, Identifier):
        if len(table_obj.parts) == 1:
            name = (default_db_name, table_obj.parts[0])
        else:
            name = (table_obj.parts[0], table_obj.parts[-1])
    elif isinstance(table_obj, Select):
        # it is subquery
        if table_obj.alias is None:
            name = 't'
        else:
            name = table_obj.alias.parts[0]
        name = (default_db_name, name)
    elif isinstance(table_obj, Join):
        # get from first table
        return get_table_alias(table_obj.left, default_db_name)
    else:
        # unknown yet object
        return default_db_name, 't', 't'

    if table_obj.alias is not None:
        name = name + ('.'.join(table_obj.alias.parts),)
    else:
        name = name + (name[1],)
    return name


def get_fill_param_fnc(steps_data):
    def fill_params(node, callstack=None, **kwargs):
        if isinstance(node, Parameter):
            rs = steps_data[node.value.step_num]
            items = [Constant(i) for i in rs.get_column_values(col_idx=0)]

            is_single_item = True
            if callstack:
                node_prev = callstack[0]
                if isinstance(node_prev, BinaryOperation):
                    # Check case: 'something IN Parameter()'
                    if node_prev.op.lower() == 'in' and node_prev.args[1] is node:
                        is_single_item = False

            if is_single_item and len(items) == 1:
                # extract one value for option 'col=(subselect)'
                node = items[0]
            else:
                node = Tuple(items)
            return node

        if isinstance(node, Parameter):
            rs = steps_data[node.value.step_num]
            items = [Constant(i) for i in rs.get_column_values(col_idx=0)]
            return Tuple(items)
    return fill_params


class FetchDataframeStepCall(BaseStepCall):

    bind = FetchDataframeStep

    def call(self, step):

        dn = self.session.datahub.get(step.integration)
        query = step.query

        if dn is None:
            raise UnknownError(f'Unknown integration name: {step.integration}')

        if query is None:
            table_alias = (self.context.get('database'), 'result', 'result')

            # fetch raw_query
            response = dn.query(
                native_query=step.raw_query,
                session=self.session
            )
            df = response.data_frame
        else:
            table_alias = get_table_alias(step.query.from_table, self.context.get('database'))

            # TODO for information_schema we have 'database' = 'mindsdb'

            # fill params
            fill_params = get_fill_param_fnc(self.steps_data)
            query_traversal(query, fill_params)

            query, context_callback = query_context_controller.handle_db_context_vars(query, dn, self.session)

            response = dn.query(
                query=query,
                session=self.session
            )
            df = response.data_frame

            if context_callback:
                context_callback(df, response.columns)

        result = ResultSet()

        result.from_df(
            df,
            table_name=table_alias[1],
            table_alias=table_alias[2],
            database=table_alias[0]
        )

        return result



import pytest
import json
from pymongo import MongoClient

from mindsdb_sql_parser.ast import Identifier, Select, Star
from mindsdb.integrations.handlers.mongodb_handler.mongodb_handler import MongoDBHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE


HANDLER_KWARGS = {
    "connection_data": {
        "host": "127.0.0.1",
        "port": "27017",
        "username": "test_user",
        "password": "supersecret",
        "database": "mongo_test_db",
    }
}

expected_columns = ["_id", "col_one", "col_two", "col_three", "col_four"]


def seed_db():
    """Seed the test DB with some data"""
    creds = HANDLER_KWARGS["connection_data"]
    uri = f"mongodb://{creds['username']}:{creds['password']}@{creds['host']}"
    conn = MongoClient(uri)
    db = conn[HANDLER_KWARGS["connection_data"]["database"]]  # noqa

    with open("mindsdb/integrations/handlers/mongodb_handler/tests/seed.json", "r") as f:
        json.load(f)
    conn.close()


@pytest.fixture(scope="module")
def handler(request):
    seed_db()
    handler = MongoDBHandler("mongo_handler", **HANDLER_KWARGS)
    return handler


def check_valid_response(res):
    if res.resp_type == RESPONSE_TYPE.TABLE:
        assert res.data_frame is not None, "expected to have some data, but got None"
    assert (
        res.error_code == 0
    ), f"expected to have zero error_code, but got {res.error_code}"
    assert (
        res.error_message is None
    ), f"expected to have None in error message, but got {res.error_message}"


""" TESTS """

# TODO - Subscribe


class TestMongoDBConnection:
    def test_connect(self, handler):
        handler.connect()
        assert handler.is_connected, "connection error"

    def test_check_connection(self, handler):
        res = handler.check_connection()
        assert res.success is True, res.error_message


# TODO - Subscribe


class TestMongoDBQuery:
    def test_native_query(self, handler):
        query_string = "db.test.find()"
        response = handler.native_query(query_string)
        dbs = response.data_frame
        assert dbs is not None, "expected to get some data, but got None"
        assert "col_one" in dbs, f"expected to get 'col_one' column in response:\n{dbs}"

    def test_select_query(self, handler):
        limit = 3
        query = Select(
            targets=[Star()],
            from_table=Identifier(parts=["test"]),
        )
        res = handler.query(query)
        check_valid_response(res)
        got_rows = res.data_frame.shape[0]
        want_rows = limit
        assert (
            got_rows == want_rows
        ), f"expected to have {want_rows} rows in response but got: {got_rows}"


class TestMongoDBTables:
    def test_get_tables(self, handler):
        res = handler.get_tables()
        tables = res.data_frame
        test_table = list(tables["table_name"])
        assert (
            tables is not None
        ), "expected to have some table_name in the db, but got None"
        assert (
            "table_name" in tables
        ), f"expected to get 'table_name' column in the response:\n{tables}"
        # get a specific table from the tables list
        assert (
            "test" in test_table
        ), f"expected to have 'test' table in the db but got: {test_table}"


class TestMongoDBColumns:
    def test_get_columns(self, handler):
        described = handler.get_columns("test")
        describe_data = described.data_frame
        check_valid_response(described)
        got_columns = list(describe_data.iloc[:, 0])
        assert (
            got_columns == expected_columns
        ), f"expected to have next columns in test table:\n{expected_columns}\nbut got:\n{got_columns}"


class TestMongoDBDisconnect:
    def test_disconnect(self, handler):
        handler.disconnect()
        assert handler.is_connected is False, "failed to disconnect"
